from collections.abc import Callable, Mapping, Sequence
from datetime import datetime
from typing import Any

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.partitions.definition.time_window import (
    TimeWindowPartitionsDefinition,
)
from dagster._core.definitions.partitions.definition.time_window_subclasses import (
    DailyPartitionsDefinition,
    HourlyPartitionsDefinition,
    MonthlyPartitionsDefinition,
    WeeklyPartitionsDefinition,
)
from dagster._core.definitions.partitions.partitioned_config import PartitionedConfig
from dagster._core.definitions.timestamp import TimestampWithTimezone


def wrap_time_window_tags_fn(
    tags_fn: Callable[[datetime, datetime], Mapping[str, str]] | None,
    partitions_def: TimeWindowPartitionsDefinition,
) -> Callable[[str], Mapping[str, str]]:
    def _tag_wrapper(key: str) -> Mapping[str, str]:
        if not tags_fn:
            return {}
        time_window = partitions_def.time_window_for_partition_key(key)
        return tags_fn(time_window.start, time_window.end)

    return _tag_wrapper


def wrap_time_window_run_config_fn(
    run_config_fn: Callable[[datetime, datetime], Mapping[str, Any]] | None,
    partitions_def: TimeWindowPartitionsDefinition,
) -> Callable[[str], Mapping[str, Any]]:
    def _run_config_wrapper(key: str) -> Mapping[str, Any]:
        if not run_config_fn:
            return {}
        time_window = partitions_def.time_window_for_partition_key(key)
        return run_config_fn(time_window.start, time_window.end)

    return _run_config_wrapper


@public
def hourly_partitioned_config(
    start_date: datetime | str,
    minute_offset: int = 0,
    timezone: str | None = None,
    fmt: str | None = None,
    end_offset: int = 0,
    tags_for_partition_fn: Callable[[datetime, datetime], Mapping[str, str]] | None = None,
    exclusions: Sequence[str | datetime | TimestampWithTimezone] | None = None,
) -> Callable[
    [Callable[[datetime, datetime], Mapping[str, Any]]],
    PartitionedConfig[HourlyPartitionsDefinition],
]:
    """Defines run config over a set of hourly partitions.

    The decorated function should accept a start datetime and end datetime, which represent the date
    partition the config should delineate.

    The decorated function should return a run config dictionary.

    The resulting object created by this decorator can be provided to the config argument of a Job.
    The first partition in the set will start at the start_date at midnight. The last partition in
    the set will end before the current time, unless the end_offset argument is set to a positive
    number. If minute_offset is provided, the start and end times of each partition will be
    minute_offset past the hour.

    Args:
        start_date (Union[datetime.datetime, str]): The first date in the set of partitions. Can
            provide in either a datetime or string format.
        minute_offset (int): Number of minutes past the hour to "split" the partition. Defaults
            to 0.
        fmt (Optional[str]): The date format to use. Defaults to `%Y-%m-%d`.
        timezone (Optional[str]): The timezone in which each date should exist.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>`_ - e.g. "America/Los_Angeles".
        end_offset (int): Extends the partition set by a number of partitions equal to the value
            passed. If end_offset is 0 (the default), the last partition ends before the current
            time. If end_offset is 1, the second-to-last partition ends before the current time,
            and so on.
        tags_for_partition_fn (Optional[Callable[[str], Mapping[str, str]]]): A function that
            accepts a partition time window and returns a dictionary of tags to attach to runs for
            that partition.
        exclusions (Optional[Sequence[Union[str, datetime]]]): Specifies a sequence of cron strings
            or datetime objects that should be excluded from the partition set. Every tick of the
            cron schedule that matches an excluded datetime or matches the tick of an excluded
            cron string will be excluded from the partition set.

    .. code-block:: python

        from datetime import datetime
        from dagster import hourly_partitioned_config

        @hourly_partitioned_config(start_date=datetime(2022, 3, 12))
        def my_hourly_partitioned_config(start: datetime, end: datetime):
            return {"start": start.strftime("%Y-%m-%d %H:%M"), "end": end.strftime("%Y-%m-%d %H:%M")}
        # creates partitions (2022-03-12-00:00, 2022-03-12-01:00), (2022-03-12-01:00, 2022-03-12-02:00), ...

        @hourly_partitioned_config(start_date=datetime(2022, 3, 12), minute_offset=15)
        def my_offset_hourly_partitioned_config(start: datetime, end: datetime):
            return {"start": start.strftime("%Y-%m-%d %H:%M"), "end": end.strftime("%Y-%m-%d %H:%M")}
        # creates partitions (2022-03-12-00:15, 2022-03-12-01:15), (2022-03-12-01:15, 2022-03-12-02:15), ...
    """

    def inner(
        fn: Callable[[datetime, datetime], Mapping[str, Any]],
    ) -> PartitionedConfig[HourlyPartitionsDefinition]:
        check.callable_param(fn, "fn")

        partitions_def = HourlyPartitionsDefinition(
            start_date=start_date,
            minute_offset=minute_offset,
            timezone=timezone,
            fmt=fmt,
            end_offset=end_offset,
            exclusions=exclusions,
        )
        return PartitionedConfig(
            run_config_for_partition_key_fn=wrap_time_window_run_config_fn(fn, partitions_def),
            partitions_def=partitions_def,
            decorated_fn=fn,
            tags_for_partition_key_fn=wrap_time_window_tags_fn(
                tags_for_partition_fn, partitions_def
            ),
        )

    return inner


@public
def daily_partitioned_config(
    start_date: datetime | str,
    minute_offset: int = 0,
    hour_offset: int = 0,
    timezone: str | None = None,
    fmt: str | None = None,
    end_offset: int = 0,
    tags_for_partition_fn: Callable[[datetime, datetime], Mapping[str, str]] | None = None,
    exclusions: Sequence[str | datetime | TimestampWithTimezone] | None = None,
) -> Callable[
    [Callable[[datetime, datetime], Mapping[str, Any]]],
    PartitionedConfig[DailyPartitionsDefinition],
]:
    """Defines run config over a set of daily partitions.

    The decorated function should accept a start datetime and end datetime, which represent the bounds
    of the date partition the config should delineate.

    The decorated function should return a run config dictionary.

    The resulting object created by this decorator can be provided to the config argument of a Job.
    The first partition in the set will start at the start_date at midnight. The last partition in
    the set will end before the current time, unless the end_offset argument is set to a positive
    number. If minute_offset and/or hour_offset are used, the start and end times of each partition
    will be hour_offset:minute_offset of each day.

    Args:
        start_date (Union[datetime.datetime, str]): The first date in the set of partitions. Can
            provide in either a datetime or string format.
        minute_offset (int): Number of minutes past the hour to "split" the partition. Defaults
            to 0.
        hour_offset (int): Number of hours past 00:00 to "split" the partition. Defaults to 0.
        timezone (Optional[str]): The timezone in which each date should exist.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>`_ - e.g. "America/Los_Angeles".
        fmt (Optional[str]): The date format to use. Defaults to `%Y-%m-%d`.
        end_offset (int): Extends the partition set by a number of partitions equal to the value
            passed. If end_offset is 0 (the default), the last partition ends before the current
            time. If end_offset is 1, the second-to-last partition ends before the current time,
            and so on.
        tags_for_partition_fn (Optional[Callable[[str], Mapping[str, str]]]): A function that
            accepts a partition time window and returns a dictionary of tags to attach to runs for
            that partition.
        exclusions (Optional[Sequence[Union[str, datetime]]]): Specifies a sequence of cron strings
            or datetime objects that should be excluded from the partition set. Every tick of the
            cron schedule that matches an excluded datetime or matches the tick of an excluded
            cron string will be excluded from the partition set.

    .. code-block:: python

        from datetime import datetime
        from dagster import daily_partitioned_config

        @daily_partitioned_config(start_date="2022-03-12")
        def my_partitioned_config(start: datetime, end: datetime):
            return {"start": start.strftime("%Y-%m-%d"), "end": end.strftime("%Y-%m-%d")}
        # creates partitions (2022-03-12-00:00, 2022-03-13-00:00), (2022-03-13-00:00, 2022-03-14-00:00), ...

        @daily_partitioned_config(start_date="2022-03-12", minute_offset=15, hour_offset=16)
        def my_offset_partitioned_config(start: datetime, end: datetime):
            return {"start": start.strftime("%Y-%m-%d"), "end": end.strftime("%Y-%m-%d")}
        # creates partitions (2022-03-12-16:15, 2022-03-13-16:15), (2022-03-13-16:15, 2022-03-14-16:15), ...
    """

    def inner(
        fn: Callable[[datetime, datetime], Mapping[str, Any]],
    ) -> PartitionedConfig[DailyPartitionsDefinition]:
        check.callable_param(fn, "fn")

        partitions_def = DailyPartitionsDefinition(
            start_date=start_date,
            minute_offset=minute_offset,
            hour_offset=hour_offset,
            timezone=timezone,
            fmt=fmt,
            end_offset=end_offset,
            exclusions=exclusions,
        )

        return PartitionedConfig(
            run_config_for_partition_key_fn=wrap_time_window_run_config_fn(fn, partitions_def),
            partitions_def=partitions_def,
            decorated_fn=fn,
            tags_for_partition_key_fn=wrap_time_window_tags_fn(
                tags_for_partition_fn, partitions_def
            ),
        )

    return inner


@public
def weekly_partitioned_config(
    start_date: datetime | str,
    minute_offset: int = 0,
    hour_offset: int = 0,
    day_offset: int = 0,
    timezone: str | None = None,
    fmt: str | None = None,
    end_offset: int = 0,
    tags_for_partition_fn: Callable[[datetime, datetime], Mapping[str, str]] | None = None,
    exclusions: Sequence[str | datetime | TimestampWithTimezone] | None = None,
) -> Callable[
    [Callable[[datetime, datetime], Mapping[str, Any]]],
    PartitionedConfig[WeeklyPartitionsDefinition],
]:
    """Defines run config over a set of weekly partitions.

    The decorated function should accept a start datetime and end datetime, which represent the date
    partition the config should delineate.

    The decorated function should return a run config dictionary.

    The resulting object created by this decorator can be provided to the config argument of a Job.
    The first partition in the set will start at the start_date. The last partition in the set will
    end before the current time, unless the end_offset argument is set to a positive number. If
    day_offset is provided, the start and end date of each partition will be day of the week
    corresponding to day_offset (0 indexed with Sunday as the start of the week). If
    minute_offset and/or hour_offset are used, the start and end times of each partition will be
    hour_offset:minute_offset of each day.

    Args:
        start_date (Union[datetime.datetime, str]): The first date in the set of partitions will
            Sunday at midnight following start_date. Can provide in either a datetime or string
            format.
        minute_offset (int): Number of minutes past the hour to "split" the partition. Defaults
            to 0.
        hour_offset (int): Number of hours past 00:00 to "split" the partition. Defaults to 0.
        day_offset (int): Day of the week to "split" the partition. Defaults to 0 (Sunday).
        timezone (Optional[str]): The timezone in which each date should exist.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>`_ - e.g. "America/Los_Angeles".
        fmt (Optional[str]): The date format to use. Defaults to `%Y-%m-%d`.
        end_offset (int): Extends the partition set by a number of partitions equal to the value
            passed. If end_offset is 0 (the default), the last partition ends before the current
            time. If end_offset is 1, the second-to-last partition ends before the current time,
            and so on.
        tags_for_partition_fn (Optional[Callable[[str], Mapping[str, str]]]): A function that
            accepts a partition time window and returns a dictionary of tags to attach to runs for
            that partition.
        exclusions (Optional[Sequence[Union[str, datetime]]]): Specifies a sequence of cron strings
            or datetime objects that should be excluded from the partition set. Every tick of the
            cron schedule that matches an excluded datetime or matches the tick of an excluded
            cron string will be excluded from the partition set.

    .. code-block:: python

        from datetime import datetime
        from dagster import weekly_partitioned_config

        @weekly_partitioned_config(start_date="2022-03-12")
        def my_weekly_partitioned_config(start: datetime, end: datetime):
            return {"start": start.strftime("%Y-%m-%d"), "end": end.strftime("%Y-%m-%d")}
        # creates partitions (2022-03-13-00:00, 2022-03-20-00:00), (2022-03-20-00:00, 2022-03-27-00:00), ...

        @weekly_partitioned_config(start_date="2022-03-12", minute_offset=15, hour_offset=3, day_offset=6)
        def my_offset_weekly_partitioned_config(start: datetime, end: datetime):
            return {"start": start.strftime("%Y-%m-%d"), "end": end.strftime("%Y-%m-%d")}
        # creates partitions (2022-03-12-03:15, 2022-03-19-03:15), (2022-03-19-03:15, 2022-03-26-03:15), ...
    """

    def inner(
        fn: Callable[[datetime, datetime], Mapping[str, Any]],
    ) -> PartitionedConfig[WeeklyPartitionsDefinition]:
        check.callable_param(fn, "fn")

        partitions_def = WeeklyPartitionsDefinition(
            start_date=start_date,
            minute_offset=minute_offset,
            hour_offset=hour_offset,
            day_offset=day_offset,
            timezone=timezone,
            fmt=fmt,
            end_offset=end_offset,
            exclusions=exclusions,
        )
        return PartitionedConfig(
            run_config_for_partition_key_fn=wrap_time_window_run_config_fn(fn, partitions_def),
            partitions_def=partitions_def,
            decorated_fn=fn,
            tags_for_partition_key_fn=wrap_time_window_tags_fn(
                tags_for_partition_fn, partitions_def
            ),
        )

    return inner


@public
def monthly_partitioned_config(
    start_date: datetime | str,
    minute_offset: int = 0,
    hour_offset: int = 0,
    day_offset: int = 1,
    timezone: str | None = None,
    fmt: str | None = None,
    end_offset: int = 0,
    tags_for_partition_fn: Callable[[datetime, datetime], Mapping[str, str]] | None = None,
    exclusions: Sequence[str | datetime | TimestampWithTimezone] | None = None,
) -> Callable[
    [Callable[[datetime, datetime], Mapping[str, Any]]],
    PartitionedConfig[MonthlyPartitionsDefinition],
]:
    """Defines run config over a set of monthly partitions.

    The decorated function should accept a start datetime and end datetime, which represent the date
    partition the config should delineate.

    The decorated function should return a run config dictionary.

    The resulting object created by this decorator can be provided to the config argument of a Job.
    The first partition in the set will start at midnight on the soonest first of the month after
    start_date. The last partition in the set will end before the current time, unless the
    end_offset argument is set to a positive number. If day_offset is provided, the start and end
    date of each partition will be day_offset. If minute_offset and/or hour_offset are used, the
    start and end times of each partition will be hour_offset:minute_offset of each day.

    Args:
        start_date (Union[datetime.datetime, str]): The first date in the set of partitions will be
            midnight the soonest first of the month following start_date. Can provide in either a
            datetime or string format.
        minute_offset (int): Number of minutes past the hour to "split" the partition. Defaults
            to 0.
        hour_offset (int): Number of hours past 00:00 to "split" the partition. Defaults to 0.
        day_offset (int): Day of the month to "split" the partition. Defaults to 1.
        timezone (Optional[str]): The timezone in which each date should exist.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>`_ - e.g. "America/Los_Angeles".
        fmt (Optional[str]): The date format to use. Defaults to `%Y-%m-%d`.
        end_offset (int): Extends the partition set by a number of partitions equal to the value
            passed. If end_offset is 0 (the default), the last partition ends before the current
            time. If end_offset is 1, the second-to-last partition ends before the current time,
            and so on.
        tags_for_partition_fn (Optional[Callable[[str], Mapping[str, str]]]): A function that
            accepts a partition time window and returns a dictionary of tags to attach to runs for
            that partition.
        exclusions (Optional[Sequence[Union[str, datetime]]]): Specifies a sequence of cron strings
            or datetime objects that should be excluded from the partition set. Every tick of the
            cron schedule that matches an excluded datetime or matches the tick of an excluded
            cron string will be excluded from the partition set.

    .. code-block:: python

        from datetime import datetime
        from dagster import monthly_partitioned_config

        @monthly_partitioned_config(start_date="2022-03-12")
        def my_monthly_partitioned_config(start: datetime, end: datetime):
            return {"start": start.strftime("%Y-%m-%d"), "end": end.strftime("%Y-%m-%d")}
        # creates partitions (2022-04-01-00:00, 2022-05-01-00:00), (2022-05-01-00:00, 2022-06-01-00:00), ...

        @monthly_partitioned_config(start_date="2022-03-12", minute_offset=15, hour_offset=3, day_offset=5)
        def my_offset_monthly_partitioned_config(start: datetime, end: datetime):
            return {"start": start.strftime("%Y-%m-%d"), "end": end.strftime("%Y-%m-%d")}
        # creates partitions (2022-04-05-03:15, 2022-05-05-03:15), (2022-05-05-03:15, 2022-06-05-03:15), ...
    """

    def inner(
        fn: Callable[[datetime, datetime], Mapping[str, Any]],
    ) -> PartitionedConfig[MonthlyPartitionsDefinition]:
        check.callable_param(fn, "fn")

        partitions_def = MonthlyPartitionsDefinition(
            start_date=start_date,
            minute_offset=minute_offset,
            hour_offset=hour_offset,
            day_offset=day_offset,
            timezone=timezone,
            fmt=fmt,
            end_offset=end_offset,
            exclusions=exclusions,
        )

        return PartitionedConfig(
            run_config_for_partition_key_fn=wrap_time_window_run_config_fn(fn, partitions_def),
            partitions_def=partitions_def,
            decorated_fn=fn,
            tags_for_partition_key_fn=wrap_time_window_tags_fn(
                tags_for_partition_fn, partitions_def
            ),
        )

    return inner



from dagster.components.lib.shim_components.base import ShimScaffolder
from dagster.components.scaffold.scaffold import ScaffoldRequest, scaffold_with


class ResourcesScaffolder(ShimScaffolder):
    def get_text(self, request: ScaffoldRequest) -> str:
        return """import dagster as dg


@dg.definitions
def resources() -> dg.Definitions:
    return dg.Definitions(resources={})
"""


def resources() -> None:
    """Symbol for dg scaffold to target."""
    ...


scaffold_with(ResourcesScaffolder)(resources)



import sys
from typing import TYPE_CHECKING, Annotated, Any, Generic, TypeVar, Union, cast, get_origin

from pydantic import Field
from typing_extensions import Self, dataclass_transform

from dagster._config.pythonic_config.type_check_utils import safe_is_subclass
from dagster._core.errors import DagsterInvalidDagsterTypeInPythonicConfigDefinitionError

try:
    # Pydantic 1.x
    from pydantic._internal._model_construction import ModelMetaclass
except ImportError:
    # Pydantic 2.x
    from pydantic.main import ModelMetaclass

if TYPE_CHECKING:
    from dagster._config.pythonic_config import PartialResource


def _materialize_annotations_for_pydantic(namespaces: dict[str, Any]) -> dict[str, Any]:
    """Materialize lazy annotations in Python 3.14+ (PEP 649)."""
    annotations = namespaces.get("__annotations__", {})

    # if annotations are empty, try calling __annotate_func__ (PEP 649)
    if sys.version_info >= (3, 14) and not annotations:
        import annotationlib

        annotate_func = annotationlib.get_annotate_from_class_namespace(namespaces)
        if annotate_func is not None:
            annotations = annotationlib.call_annotate_function(
                annotate_func, annotationlib.Format.VALUE
            )
            namespaces["__annotations__"] = annotations

    return annotations


# Since a metaclass is invoked by Resource before Resource or PartialResource is defined, we need to
# define a temporary class to use as a placeholder for use in the initial metaclass invocation.
#
# These initial invocations will use the placeholder values, which is fine, since there's no
# attributes on the Resource class which would be affected. The only time the metaclass will
# actually change the type annotations is when it's invoked for a user-created subclass of Resource,
# at which point the placeholder values will be replaced with the actual types.
class LateBoundTypesForResourceTypeChecking:
    _TResValue = TypeVar("_TResValue")

    class _Temp(Generic[_TResValue]):
        pass

    _ResourceDep: type = _Temp
    _Resource: type = _Temp
    _PartialResource: type = _Temp

    @staticmethod
    def get_resource_rep_type() -> type:
        return LateBoundTypesForResourceTypeChecking._ResourceDep

    @staticmethod
    def get_resource_type() -> type:
        return LateBoundTypesForResourceTypeChecking._Resource

    @staticmethod
    def get_partial_resource_type(base: type) -> type:
        # LateBoundTypesForResourceTypeChecking._PartialResource[base] would be the more
        # correct thing to return, but to enable that deeper pydantic integration
        # needs to be done on the PartialResource class
        # https://github.com/dagster-io/dagster/issues/18017
        return LateBoundTypesForResourceTypeChecking._PartialResource

    @staticmethod
    def set_actual_types_for_type_checking(
        resource_dep_type: type, resource_type: type, partial_resource_type: type
    ) -> None:
        LateBoundTypesForResourceTypeChecking._ResourceDep = resource_dep_type
        LateBoundTypesForResourceTypeChecking._Resource = resource_type
        LateBoundTypesForResourceTypeChecking._PartialResource = partial_resource_type


@dataclass_transform(kw_only_default=True, field_specifiers=(Field,))
class BaseConfigMeta(ModelMetaclass):  # type: ignore
    def __new__(cls, name, bases, namespaces, **kwargs) -> Any:
        # Materialize lazy annotations for Python 3.14+ before Pydantic sees them
        annotations = _materialize_annotations_for_pydantic(namespaces)

        # Need try/catch because DagsterType may not be loaded when some of the base Config classes are
        # being created
        # Any user-created Config class will have DagsterType loaded by the time it's created, so this
        # will only affect the base Config classes (where this error won't be an issue)
        try:
            from dagster._core.types.dagster_type import DagsterType

            for field in annotations:
                if isinstance(annotations[field], DagsterType):
                    raise DagsterInvalidDagsterTypeInPythonicConfigDefinitionError(name, field)

        except ImportError:
            pass

        return super().__new__(cls, name, bases, namespaces, **kwargs)


@dataclass_transform(kw_only_default=True, field_specifiers=(Field,))
class BaseResourceMeta(BaseConfigMeta):
    """Custom metaclass for Resource and PartialResource. This metaclass is responsible for
    transforming the type annotations on the class so that Pydantic constructor-time validation
    does not error when users provide partially configured resources to resource params.

    For example, the following code would ordinarily fail Pydantic validation:

    .. code-block:: python

        class FooResource(ConfigurableResource):
            bar: BarResource

        # Types as PartialResource[BarResource]
        partial_bar = BarResource.configure_at_runtime()

        # Pydantic validation fails because bar is not a BarResource
        foo = FooResource(bar=partial_bar)

    This metaclass transforms the type annotations on the class so that Pydantic validation
    accepts either a PartialResource or a Resource as a value for the resource dependency.
    """

    def __new__(cls, name, bases, namespaces, **kwargs) -> Any:
        from pydantic.fields import FieldInfo

        # Gather all type annotations from the class and its base classes
        # Materialize lazy annotations for Python 3.14+ before Pydantic sees them
        annotations = _materialize_annotations_for_pydantic(namespaces)

        for field in annotations:
            if not field.startswith("__"):
                # Check if the annotation is a ResourceDependency
                if (
                    get_origin(annotations[field])
                    == LateBoundTypesForResourceTypeChecking.get_resource_rep_type()
                ):
                    # arg = get_args(annotations[field])[0]
                    # If so, we treat it as a Union of a PartialResource and a Resource
                    # for Pydantic's sake.
                    annotations[field] = Annotated[Any, "resource_dependency"]
                elif safe_is_subclass(
                    annotations[field], LateBoundTypesForResourceTypeChecking.get_resource_type()
                ):
                    # If the annotation is a Resource, we treat it as a Union of a PartialResource
                    # and a Resource for Pydantic's sake, so that a user can pass in a partially
                    # configured resource.
                    base = annotations[field]
                    annotations[field] = Annotated[
                        base
                        | LateBoundTypesForResourceTypeChecking.get_partial_resource_type(base),
                        "resource_dependency",
                    ]
                    # Pydantic 2.5.0 changed the default union mode to "smart", which causes
                    # partial resource initialization to fail, since Pydantic would attempt to
                    # initialize a PartialResource with parameters which are invalid.
                    # https://github.com/pydantic/pydantic-core/pull/867
                    # Here, we explicitly set the union mode to "left_to_right".
                    # https://docs.pydantic.dev/latest/concepts/unions/#left-to-right-mode
                    namespaces[field] = FieldInfo(
                        union_mode="left_to_right", annotation=annotations[field]
                    )

        namespaces["__annotations__"] = annotations
        return super().__new__(cls, name, bases, namespaces, **kwargs)


class TypecheckAllowPartialResourceInitParams:
    """Implementation of the Python descriptor protocol (https://docs.python.org/3/howto/descriptor.html)
    to adjust the types of resource inputs and outputs, e.g. resource dependencies can be passed in
    as PartialResources or Resources, but will always be returned as Resources.

    For example, given a resource with the following signature:

    .. code-block:: python

        class FooResource(Resource):
            bar: BarResource

    The following code will work:

    .. code-block:: python

        # Types as PartialResource[BarResource]
        partial_bar = BarResource.configure_at_runtime()

        # bar parameter takes BarResource | PartialResource[BarResource]
        foo = FooResource(bar=partial_bar)

        # initialization of FooResource succeeds,
        # populating the bar attribute with a full BarResource

        # bar attribute is typed as BarResource, since
        # it is fully initialized when a user accesses it
        print(foo.bar)

    Very similar to https://github.com/pydantic/pydantic/discussions/4262.
    """

    def __set_name__(self, _owner, name):
        self._assigned_name = name

    def __get__(self: Self, obj: Any, owner: Any) -> Self:
        # no-op implementation (only used to affect type signature)
        return cast("Self", getattr(obj, self._assigned_name))

    # The annotation her has been temporarily changed from:
    #     value: Union[Self, "PartialResource[Self]"]
    # to:
    #     value: Union[Any, "PartialResource[Any]"]
    # This is because of a bug in mypy that is incorrectly interpreting the
    # signature and can cause a false positive type error for users. This only
    # started being detected in our test_type_signatures.py tests on 2024-02-02
    # when some annotations elsewhere were added, likely causing mypy to
    # analyze code it was previously skipping. The annotation should be
    # reverted when the bug is fixed or another solution that surface as type
    # errors for mypy users is found.
    def __set__(self, obj: object | None, value: Union[Any, "PartialResource[Any]"]) -> None:
        # no-op implementation (only used to affect type signature)
        setattr(obj, self._assigned_name, value)



"""Run methods implementation - consolidated from RunsMixin and RunDomain."""

from collections.abc import Mapping, Sequence, Set
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional

from dagster._core.definitions.events import AssetKey
from dagster._core.storage.dagster_run import (
    DagsterRun,
    DagsterRunStatus,
    JobBucket,
    RunPartitionData,
    RunsFilter,
    TagBucket,
)
from dagster._utils import traced

if TYPE_CHECKING:
    from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey
    from dagster._core.definitions.assets.graph.base_asset_graph import BaseAssetGraph
    from dagster._core.definitions.job_definition import JobDefinition
    from dagster._core.definitions.repository_definition.repository_definition import (
        RepositoryLoadData,
    )
    from dagster._core.events import DagsterEvent
    from dagster._core.execution.plan.plan import ExecutionPlan
    from dagster._core.execution.plan.resume_retry import ReexecutionStrategy
    from dagster._core.execution.stats import RunStepKeyStatsSnapshot
    from dagster._core.instance.instance import DagsterInstance
    from dagster._core.instance.runs.run_domain import RunDomain
    from dagster._core.origin import JobPythonOrigin
    from dagster._core.remote_origin import RemoteJobOrigin
    from dagster._core.remote_representation.code_location import CodeLocation
    from dagster._core.remote_representation.external import RemoteJob
    from dagster._core.snap import ExecutionPlanSnapshot, JobSnap
    from dagster._core.storage.dagster_run import DagsterRunStatsSnapshot
    from dagster._core.storage.event_log import EventLogStorage
    from dagster._core.storage.runs import RunStorage
    from dagster._core.workspace.context import BaseWorkspaceRequestContext


class RunMethods:
    """Mixin class containing run-related functionality for DagsterInstance.

    This class consolidates run operations from RunsMixin and provides the run_domain
    property previously in DomainsMixin.
    """

    @property
    def _instance(self) -> "DagsterInstance":
        """Cast self to DagsterInstance for type-safe access to instance methods and properties."""
        import dagster._check as check
        from dagster._core.instance.instance import DagsterInstance

        return check.inst(self, DagsterInstance)

    # Private member access wrappers
    @property
    def _run_storage_impl(self) -> "RunStorage":
        """Access to run storage."""
        return self._instance._run_storage  # noqa: SLF001

    @property
    def _event_storage_impl(self) -> "EventLogStorage":
        """Access to event storage."""
        return self._instance._event_storage  # noqa: SLF001

    # Domain property - moved from DomainsMixin

    @cached_property
    def run_domain(self) -> "RunDomain":
        """Get run domain for complex business logic."""
        from dagster._core.instance.runs.run_domain import RunDomain

        return RunDomain(self._instance)

    # Core Run Retrieval Methods - moved from RunsMixin

    @traced
    def get_run_stats(self, run_id: str) -> "DagsterRunStatsSnapshot":
        return self._event_storage_impl.get_stats_for_run(run_id)

    @traced
    def get_run_step_stats(
        self, run_id: str, step_keys: Sequence[str] | None = None
    ) -> Sequence["RunStepKeyStatsSnapshot"]:
        return self._event_storage_impl.get_step_stats_for_run(run_id, step_keys)

    @traced
    def get_run_tags(
        self,
        tag_keys: Sequence[str],
        value_prefix: str | None = None,
        limit: int | None = None,
    ) -> Sequence[tuple[str, set[str]]]:
        return self._run_storage_impl.get_run_tags(
            tag_keys=tag_keys, value_prefix=value_prefix, limit=limit
        )

    @traced
    def get_run_tag_keys(self) -> Sequence[str]:
        return self._run_storage_impl.get_run_tag_keys()

    @traced
    def get_run_group(self, run_id: str) -> tuple[str, Sequence[DagsterRun]] | None:
        return self._run_storage_impl.get_run_group(run_id)

    # Run Creation Methods - moved from RunsMixin

    def create_run_for_job(
        self,
        job_def: "JobDefinition",
        execution_plan: Optional["ExecutionPlan"] = None,
        run_id: str | None = None,
        run_config: Mapping[str, object] | None = None,
        resolved_op_selection: Set[str] | None = None,
        status: DagsterRunStatus | str | None = None,
        tags: Mapping[str, str] | None = None,
        root_run_id: str | None = None,
        parent_run_id: str | None = None,
        op_selection: Sequence[str] | None = None,
        asset_selection: Set[AssetKey] | None = None,
        remote_job_origin: Optional["RemoteJobOrigin"] = None,
        job_code_origin: Optional["JobPythonOrigin"] = None,
        repository_load_data: Optional["RepositoryLoadData"] = None,
    ) -> DagsterRun:
        """Delegate to run domain."""
        return self.run_domain.create_run_for_job(
            job_def=job_def,
            execution_plan=execution_plan,
            run_id=run_id,
            run_config=run_config,
            resolved_op_selection=resolved_op_selection,
            status=DagsterRunStatus(status) if status else None,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            op_selection=op_selection,
            asset_selection=asset_selection,
            remote_job_origin=remote_job_origin,
            job_code_origin=job_code_origin,
            repository_load_data=repository_load_data,
        )

    def create_run(
        self,
        *,
        job_name: str,
        run_id: str | None,
        run_config: Mapping[str, object] | None,
        status: DagsterRunStatus | None,
        tags: Mapping[str, Any] | None,
        root_run_id: str | None,
        parent_run_id: str | None,
        step_keys_to_execute: Sequence[str] | None,
        execution_plan_snapshot: Optional["ExecutionPlanSnapshot"],
        job_snapshot: Optional["JobSnap"],
        parent_job_snapshot: Optional["JobSnap"],
        asset_selection: Set[AssetKey] | None,
        asset_check_selection: Set["AssetCheckKey"] | None,
        resolved_op_selection: Set[str] | None,
        op_selection: Sequence[str] | None,
        remote_job_origin: Optional["RemoteJobOrigin"],
        job_code_origin: Optional["JobPythonOrigin"],
        asset_graph: "BaseAssetGraph",
    ) -> DagsterRun:
        """Create a run with the given parameters."""
        return self.run_domain.create_run(
            job_name=job_name,
            run_id=run_id,
            run_config=run_config,
            status=status,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            step_keys_to_execute=step_keys_to_execute,
            execution_plan_snapshot=execution_plan_snapshot,
            job_snapshot=job_snapshot,
            parent_job_snapshot=parent_job_snapshot,
            asset_selection=asset_selection,
            asset_check_selection=asset_check_selection,
            resolved_op_selection=resolved_op_selection,
            op_selection=op_selection,
            remote_job_origin=remote_job_origin,
            job_code_origin=job_code_origin,
            asset_graph=asset_graph,
        )

    def create_reexecuted_run(
        self,
        *,
        request_context: "BaseWorkspaceRequestContext",
        parent_run: DagsterRun,
        code_location: "CodeLocation",
        remote_job: "RemoteJob",
        strategy: "ReexecutionStrategy",
        extra_tags: Mapping[str, Any] | None = None,
        run_config: Mapping[str, Any] | None = None,
        use_parent_run_tags: bool = False,
    ) -> DagsterRun:
        return self.run_domain.create_reexecuted_run(
            request_context=request_context,
            parent_run=parent_run,
            code_location=code_location,
            remote_job=remote_job,
            strategy=strategy,
            extra_tags=extra_tags,
            run_config=run_config,
            use_parent_run_tags=use_parent_run_tags,
        )

    def register_managed_run(
        self,
        job_name: str,
        run_id: str,
        run_config: Mapping[str, object] | None,
        resolved_op_selection: Set[str] | None,
        step_keys_to_execute: Sequence[str] | None,
        tags: Mapping[str, str],
        root_run_id: str | None,
        parent_run_id: str | None,
        job_snapshot: Optional["JobSnap"],
        execution_plan_snapshot: Optional["ExecutionPlanSnapshot"],
        parent_job_snapshot: Optional["JobSnap"],
        op_selection: Sequence[str] | None = None,
        job_code_origin: Optional["JobPythonOrigin"] = None,
    ) -> DagsterRun:
        return self.run_domain.register_managed_run(
            job_name=job_name,
            run_id=run_id,
            run_config=run_config,
            resolved_op_selection=resolved_op_selection,
            step_keys_to_execute=step_keys_to_execute,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            job_snapshot=job_snapshot,
            execution_plan_snapshot=execution_plan_snapshot,
            parent_job_snapshot=parent_job_snapshot,
            op_selection=op_selection,
            job_code_origin=job_code_origin,
        )

    # Run Management Methods - moved from RunsMixin

    @traced
    def add_run(self, dagster_run: DagsterRun) -> DagsterRun:
        return self._run_storage_impl.add_run(dagster_run)

    @traced
    def handle_run_event(
        self, run_id: str, event: "DagsterEvent", update_timestamp: datetime | None = None
    ) -> None:
        return self._run_storage_impl.handle_run_event(run_id, event, update_timestamp)

    @traced
    def add_run_tags(self, run_id: str, new_tags: Mapping[str, str]) -> None:
        return self._run_storage_impl.add_run_tags(run_id, new_tags)

    @traced
    def has_run(self, run_id: str) -> bool:
        return self._run_storage_impl.has_run(run_id)

    @traced
    def get_runs(
        self,
        filters: RunsFilter | None = None,
        cursor: str | None = None,
        limit: int | None = None,
        bucket_by: JobBucket | TagBucket | None = None,
        ascending: bool = False,
    ) -> Sequence[DagsterRun]:
        return self._run_storage_impl.get_runs(filters, cursor, limit, bucket_by, ascending)

    @traced
    def get_run_ids(
        self,
        filters: RunsFilter | None = None,
        cursor: str | None = None,
        limit: int | None = None,
    ) -> Sequence[str]:
        return self._run_storage_impl.get_run_ids(filters, cursor=cursor, limit=limit)

    @traced
    def get_runs_count(self, filters: RunsFilter | None = None) -> int:
        return self._run_storage_impl.get_runs_count(filters)

    @traced
    def get_run_partition_data(self, runs_filter: RunsFilter) -> Sequence[RunPartitionData]:
        """Get run partition data for a given partitioned job."""
        return self._run_storage_impl.get_run_partition_data(runs_filter)



from abc import ABC
from collections.abc import Iterable, Mapping
from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import TYPE_CHECKING, Any, Optional

from dagster_shared.serdes.utils import SerializableTimeDelta

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.loader import LoadableBy, LoadingContext
from dagster._record import IHaveNew, record, record_custom
from dagster._serdes import deserialize_value, whitelist_for_serdes
from dagster._symbol_annotations.public import public
from dagster._time import datetime_from_timestamp, get_timezone
from dagster._utils import check
from dagster._utils.schedules import get_smallest_cron_interval, is_valid_cron_string

if TYPE_CHECKING:
    from dagster._core.events.log import EventLogEntry


@whitelist_for_serdes
class FreshnessState(str, Enum):
    PASS = "PASS"
    WARN = "WARN"
    FAIL = "FAIL"
    UNKNOWN = "UNKNOWN"
    NOT_APPLICABLE = "NOT_APPLICABLE"


@whitelist_for_serdes
@record
class FreshnessStateEvaluation:
    key: AssetKey
    freshness_state: FreshnessState


@whitelist_for_serdes
@record
class FreshnessStateChange:
    """Event that is emitted when the freshness state of an asset changes."""

    key: AssetKey
    new_state: FreshnessState
    previous_state: FreshnessState
    state_change_timestamp: float


INTERNAL_FRESHNESS_POLICY_METADATA_KEY = "dagster/internal_freshness_policy"


@public
class FreshnessPolicy(ABC):
    """Base class for all freshness policies.

    A freshness policy allows you to define expectations for the timing and frequency of asset materializations.

    An asset with a defined freshness policy can take on different freshness states:
    - ``PASS``: The asset is passing its freshness policy.
    - ``WARN``: The asset is close to failing its freshness policy.
    - ``FAIL``: The asset is failing its freshness policy.
    - ``UNKNOWN``: The asset has no materialization events, and the freshness state cannot be determined.

    If an asset does not have a freshness policy defined, it will have a freshness state of ``NOT_APPLICABLE``.

    This class provides static constructors for each of the supported freshness policy types. It is preferred to use these constructors to instantiate freshness policies, instead of instantiating the policy subtypes directly.
    """

    @classmethod
    def from_asset_spec_metadata(cls, metadata: Mapping[str, Any]) -> Optional["FreshnessPolicy"]:
        serialized_policy = metadata.get(INTERNAL_FRESHNESS_POLICY_METADATA_KEY)

        # We had a few asset spec metadatas with internal freshness policies set to literal "null" string,
        # need special handling for those cases.
        # https://github.com/dagster-io/dagster/pull/30615
        if serialized_policy is None or serialized_policy.value == "null":
            return None
        return deserialize_value(serialized_policy.value, cls)  # pyright: ignore

    @staticmethod
    def time_window(
        fail_window: timedelta, warn_window: timedelta | None = None
    ) -> "TimeWindowFreshnessPolicy":
        """Defines freshness with reference to a time window.

        Args:
            fail_window: a timedelta that defines the failure window for the asset.
            warn_window: an optional timedelta that defines the warning window for the asset.

        Returns:
            A ``TimeWindowFreshnessPolicy`` instance.

        Examples:
            .. code-block:: python

                policy = FreshnessPolicy.time_window(
                    fail_window=timedelta(hours=24), warn_window=timedelta(hours=12)
                )

            This policy expects the asset to materialize at least once every 24 hours, and warns if the latest materialization is older than 12 hours.

            - If it has been less than 12 hours since the latest materialization, the asset is passing its freshness policy, and will have a freshness state of ``PASS``.
            - If it has been between 12 and 24 hours since the latest materialization, the asset will have a freshness state of ``WARN``.
            - If it has been more than 24 hours since the latest materialization, the asset is failing its freshness policy, and will have a freshness state of ``FAIL``.

        """
        return TimeWindowFreshnessPolicy.from_timedeltas(fail_window, warn_window)

    @staticmethod
    def cron(
        deadline_cron: str, lower_bound_delta: timedelta, timezone: str = "UTC"
    ) -> "CronFreshnessPolicy":
        """Defines freshness with reference to a predetermined cron schedule.

        Args:
            deadline_cron: a cron string that defines a deadline for the asset to be materialized.
            lower_bound_delta: a timedelta that defines the lower bound for when the asset could have been materialized.
                If a deadline cron tick has passed and the most recent materialization is older than (deadline cron tick timestamp - lower bound delta),
                the asset is considered stale until it materializes again.
            timezone: optionally provide a timezone for cron evaluation. IANA time zone strings are supported. If not provided, defaults to UTC.

        Returns:
            A ``CronFreshnessPolicy`` instance.

        Examples:
            .. code-block:: python

                policy = FreshnessPolicy.cron(
                    deadline_cron="0 10 * * *",  # 10am daily
                    lower_bound_delta=timedelta(hours=1),
                )

            This policy expects the asset to materialize every day between 9:00 AM and 10:00 AM.

            - If the asset is materialized at 9:30 AM, the asset is passing its freshness policy, and will have a freshness state of ``PASS``. The asset will continue to pass the freshness policy until at least the deadline next day (10AM).
            - If the asset is materialized at 9:59 AM, the asset is passing its freshness policy, and will have a freshness state of ``PASS``. The asset will continue to pass the freshness policy until at least the deadline next day (10AM).
            - If the asset is not materialized by 10:00 AM, the asset is failing its freshness policy, and will have a freshness state of ``FAIL``. The asset will continue to fail the freshness policy until it is materialized again.
            - If the asset is then materialized at 10:30AM, it will pass the freshness policy again until at least the deadline the next day (10AM).

            Keep in mind that the policy will always look at the last completed cron tick.
            So in the example above, if asset freshness is evaluated at 9:59 AM, the policy will still consider the previous day's 9-10AM window.


        """
        return CronFreshnessPolicy(
            deadline_cron=deadline_cron, lower_bound_delta=lower_bound_delta, timezone=timezone
        )


@whitelist_for_serdes
@record
class TimeWindowFreshnessPolicy(FreshnessPolicy, IHaveNew):
    fail_window: SerializableTimeDelta
    warn_window: SerializableTimeDelta | None = None

    @classmethod
    def from_timedeltas(cls, fail_window: timedelta, warn_window: timedelta | None = None):
        check.invariant(
            fail_window.total_seconds() >= 60,
            "Due to Dagster system constraints, fail_window cannot be less than 1 minute",
        )
        if warn_window:
            check.invariant(
                warn_window.total_seconds() >= 60,
                "Due to Dagster system constraints, warn_window cannot be less than 1 minute",
            )
            check.invariant(warn_window < fail_window, "warn_window must be less than fail_window")

        return cls(
            fail_window=SerializableTimeDelta.from_timedelta(fail_window),
            warn_window=SerializableTimeDelta.from_timedelta(warn_window) if warn_window else None,
        )


@whitelist_for_serdes
@record_custom(
    field_to_new_mapping={
        "serializable_lower_bound_delta": "lower_bound_delta",
    }
)
class CronFreshnessPolicy(FreshnessPolicy, IHaveNew):
    deadline_cron: str
    serializable_lower_bound_delta: SerializableTimeDelta
    timezone: str

    def __new__(
        cls,
        deadline_cron: str,
        lower_bound_delta: timedelta | SerializableTimeDelta,
        timezone: str = "UTC",
    ):
        check.str_param(deadline_cron, "deadline_cron")
        check.invariant(is_valid_cron_string(deadline_cron), "Invalid cron string.")

        # Handle both construction (with timedelta) and deserialization (with SerializableTimeDelta)
        if isinstance(lower_bound_delta, SerializableTimeDelta):
            # During deserialization, we already have a SerializableTimeDelta
            serializable_lower_bound_delta = lower_bound_delta
            actual_lower_bound_delta = lower_bound_delta.to_timedelta()
        else:
            # During normal construction, we have a timedelta and need to convert it
            actual_lower_bound_delta = lower_bound_delta
            serializable_lower_bound_delta = SerializableTimeDelta.from_timedelta(lower_bound_delta)

        check.invariant(
            actual_lower_bound_delta >= timedelta(minutes=1),
            f"lower_bound_delta must be greater than or equal to 1 minute, was {actual_lower_bound_delta.total_seconds()} seconds",
        )
        smallest_cron_interval = get_smallest_cron_interval(deadline_cron)
        check.invariant(
            actual_lower_bound_delta <= smallest_cron_interval,
            f"lower_bound_delta must be less than or equal to the smallest cron interval of ({smallest_cron_interval.total_seconds()} seconds for deadline_cron {deadline_cron}). Provided lower_bound_delta is {actual_lower_bound_delta.total_seconds()} seconds",
        )

        try:
            get_timezone(timezone)
        # Would be better to catch a specific exception type here,
        # but it's more complicated because we use different timezone libraries
        # depending on the Python version.
        except Exception:
            raise check.CheckError(f"Invalid IANA timezone: {timezone}")
        return super().__new__(
            cls,
            deadline_cron=deadline_cron,
            serializable_lower_bound_delta=serializable_lower_bound_delta,
            timezone=timezone,
        )

    @property
    def lower_bound_delta(self) -> timedelta:
        """Returns the lower bound delta as a timedelta.
        Use this instead of accessing the serializable_lower_bound_delta directly.
        """
        return SerializableTimeDelta.to_timedelta(self.serializable_lower_bound_delta)


@whitelist_for_serdes
@record
class FreshnessStateRecordBody:
    """Store serialized metadata about the freshness state for an entity.

    Left blank for now, a few examples of what we might want to store here:
    - Source timestamp for external assets / freshness checks
    - Snapshot of the freshness policy at the time of record creation
    """

    metadata: dict[str, Any] | None


@record
class FreshnessStateRecord(LoadableBy[AssetKey]):
    entity_key: AssetKey
    freshness_state: FreshnessState
    updated_at: datetime
    record_body: FreshnessStateRecordBody

    @staticmethod
    def from_db_row(db_row):
        return FreshnessStateRecord(
            entity_key=check.not_none(AssetKey.from_db_string(db_row.entity_key)),
            freshness_state=FreshnessState(db_row.freshness_state),
            record_body=deserialize_value(db_row.record_body, FreshnessStateRecordBody),
            updated_at=db_row.update_timestamp.replace(tzinfo=timezone.utc),
        )

    @staticmethod
    def from_event_log_entry(entry: "EventLogEntry") -> "FreshnessStateRecord":
        dagster_event = check.not_none(entry.dagster_event)
        event_data = check.inst(dagster_event.event_specific_data, FreshnessStateChange)
        return FreshnessStateRecord(
            entity_key=event_data.key,
            freshness_state=event_data.new_state,
            updated_at=datetime_from_timestamp(event_data.state_change_timestamp),
            # note: metadata is not currently stored in the event log
            record_body=FreshnessStateRecordBody(metadata={}),
        )

    @classmethod
    def _blocking_batch_load(
        cls, keys: Iterable[AssetKey], context: LoadingContext
    ) -> Iterable[Optional["FreshnessStateRecord"]]:
        keys = list(keys)
        state_records = context.instance.get_freshness_state_records(keys)
        return [state_records.get(key) for key in keys]



import logging
import os
import threading
from collections.abc import Mapping, Sequence
from contextlib import AbstractContextManager, ExitStack
from typing import TYPE_CHECKING, Any, cast

from typing_extensions import Self

import dagster._check as check
from dagster._core.errors import DagsterUserCodeUnreachableError
from dagster._core.instance import DagsterInstance
from dagster._core.remote_origin import (
    GrpcServerCodeLocationOrigin,
    ManagedGrpcPythonEnvCodeLocationOrigin,
)
from dagster._core.remote_representation.grpc_server_registry import (
    GrpcServerRegistry,
    ServerRegistryEntry,
)
from dagster._core.workspace.context import WEBSERVER_GRPC_SERVER_HEARTBEAT_TTL
from dagster._core.workspace.load_target import WorkspaceLoadTarget
from dagster._grpc.constants import INCREASE_TIMEOUT_DAGSTER_YAML_MSG, GrpcServerCommand

if TYPE_CHECKING:
    from dagster._grpc.server import GrpcServerProcess


def get_auto_restart_code_server_interval() -> int:
    return int(os.getenv("DAGSTER_CODE_SERVER_AUTO_RESTART_INTERVAL", "5"))


class ProxyServerManager(AbstractContextManager):
    """Context manager that manages the lifecycle of a set of proxy code servers targeting a passed-in load target."""

    def __init__(
        self,
        instance: DagsterInstance,
        workspace_load_target: WorkspaceLoadTarget,
        code_server_log_level: str = "INFO",
    ) -> None:
        self._stack = ExitStack()

        self._instance = check.inst_param(instance, "instance", DagsterInstance)
        self._workspace_load_target = check.inst_param(
            workspace_load_target, "workspace_load_target", WorkspaceLoadTarget
        )
        self._origins = cast(
            "Sequence[GrpcServerCodeLocationOrigin | ManagedGrpcPythonEnvCodeLocationOrigin]",
            self._workspace_load_target.create_origins(),
        )

        self._grpc_server_registry: GrpcServerRegistry = self._stack.enter_context(
            GrpcServerRegistry(
                instance_ref=self._instance.get_ref(),
                server_command=GrpcServerCommand.CODE_SERVER_START,
                heartbeat_ttl=WEBSERVER_GRPC_SERVER_HEARTBEAT_TTL,
                startup_timeout=instance.code_server_process_startup_timeout,
                log_level=code_server_log_level,
                wait_for_processes_on_shutdown=instance.wait_for_local_code_server_processes_on_shutdown,
                additional_timeout_msg=INCREASE_TIMEOUT_DAGSTER_YAML_MSG,
            )
        )

        self.__shutdown_event = threading.Event()

        self._process_monitoring_threads = []
        self._initialize_endpoints()

    def _initialize_endpoints(self) -> None:
        """Initialize the endpoints for the code server processes."""
        for origin in self._origins:
            if isinstance(origin, ManagedGrpcPythonEnvCodeLocationOrigin):
                # Calling get_grpc_server_entry will start the server process if it is not already running.
                server_entry = self._grpc_server_registry.get_grpc_server_entry(origin)
                if isinstance(server_entry, ServerRegistryEntry):
                    server_process = server_entry.process
                    monitoring_thread = threading.Thread(
                        target=self._process_monitoring_thread, args=(server_process,), daemon=True
                    )
                    monitoring_thread.start()
                    self._process_monitoring_threads.append(monitoring_thread)

    def _process_monitoring_thread(self, process_entry: "GrpcServerProcess") -> None:
        """Thread responsible for monitoring the code server processes.
        - If the proxy servers exit unexpectedly, restarts them.
        - Heartbeats the proxy servers to let them know that the caller process is still alive.
        """
        while True:
            self.__shutdown_event.wait(get_auto_restart_code_server_interval())
            if self.__shutdown_event.is_set():
                break
            if process_entry.server_process and process_entry.server_process.poll() is not None:
                logging.getLogger(__name__).warning(
                    f"Code server process has exited with code {process_entry.server_process.poll()}. Restarting the code server process."
                )
                process_entry.start_server_process()
                continue
            client = process_entry.create_client()
            try:
                client.heartbeat("ping")
            except DagsterUserCodeUnreachableError:
                continue

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        self.__shutdown_event.set()
        for thread in self._process_monitoring_threads:
            thread.join()
        self._stack.close()

    def get_code_server_specs(self) -> Sequence[Mapping[str, Mapping[str, Any]]]:
        result = []
        for origin in self._origins:
            if isinstance(origin, ManagedGrpcPythonEnvCodeLocationOrigin):
                grpc_endpoint = self._grpc_server_registry.get_grpc_endpoint(origin)
                server_spec = {
                    "location_name": origin.location_name,
                    "socket": grpc_endpoint.socket,
                    "port": grpc_endpoint.port,
                    "host": grpc_endpoint.host,
                    "additional_metadata": origin.loadable_target_origin.as_dict,
                }
            else:
                server_spec = {
                    "location_name": origin.location_name,
                    "host": origin.host,
                    "port": origin.port,
                    "socket": origin.socket,
                    "additional_metadata": origin.additional_metadata,
                }
            result.append({"grpc_server": {k: v for k, v in server_spec.items() if v is not None}})
        return result



"""Utilities for discovering project paths."""

import importlib
import re
from functools import cache
from pathlib import Path

from dagster_shared import check
from dagster_shared.utils.config import (
    discover_config_file,
    get_canonical_defs_module_name,
    load_toml_as_dict,
)

from dagster._core.errors import DagsterError

_LOCAL_DEFS_STATE_DIR = ".local_defs_state"


@cache
def get_defs_module_path_for_project(project_root: Path) -> Path:
    """Get the defs module path for a project by discovering and parsing its config.

    Args:
        project_root: The root directory of the project.

    Returns:
        The filesystem path to the defs module directory.

    Raises:
        DagsterError: If no config file is found or the config is invalid.
    """
    # Discover the config file from the project root
    config_file = discover_config_file(project_root)
    if not config_file:
        raise DagsterError(f"Could not find dg.toml or pyproject.toml in {project_root}")

    # Load the config to get the root_module and defs_module
    config_dict = load_toml_as_dict(config_file)

    # Check if it's a nested config (pyproject.toml) or root (dg.toml)
    if config_file.name == "dg.toml":
        project_config = config_dict.get("project", {})
    else:
        project_config = config_dict.get("tool", {}).get("dg", {}).get("project", {})

    root_module = project_config.get("root_module")
    defs_module_name = project_config.get("defs_module")

    check.invariant(
        defs_module_name or root_module,
        f"Either defs_module or root_module must be set in the project config {config_file}",
    )

    defs_module_name = get_canonical_defs_module_name(defs_module_name, root_module)

    # Import the module to get its filesystem path
    defs_module = importlib.import_module(defs_module_name)

    # Get the module path, handling both __file__ and __path__ attributes
    if defs_module.__file__:
        # Get the directory path (handle both __init__.py and direct module files)
        defs_module_path = Path(defs_module.__file__).parent
    elif hasattr(defs_module, "__path__") and defs_module.__path__:
        defs_module_path = Path(defs_module.__path__[0])
    else:
        raise DagsterError(f"Could not determine path for module {defs_module_name}")

    return defs_module_path


def get_local_defs_state_dir(project_root: Path) -> Path:
    """Get the local state directory for the project.

    This discovers the defs module path from the project configuration and creates
    a project-local subdirectory within it for storing component state.

    Args:
        project_root: The root directory of the project.

    Returns:
        The path to the state directory.
    """
    defs_module_path = get_defs_module_path_for_project(project_root)

    local_state_dir = defs_module_path / _LOCAL_DEFS_STATE_DIR
    local_state_dir.mkdir(parents=True, exist_ok=True)

    # create .gitignore if it doesn't exist to ensure that this directory is ignored by git
    gitignore_path = local_state_dir / ".gitignore"
    if not gitignore_path.exists():
        gitignore_path.write_text("\n".join(["*", "*/", ".*"]))

    return local_state_dir


def get_local_state_dir(key: str, project_root: Path) -> Path:
    """Get the local state directory for a specific component state key.

    Args:
        key: The defs state key for the component.
        project_root: The root directory of the project.

    Returns:
        The path to the state directory for this key.
    """
    sanitized_key = re.sub(r"[^A-Za-z0-9._-]", "__", key)
    state_dir = get_local_defs_state_dir(project_root) / sanitized_key
    state_dir.mkdir(parents=True, exist_ok=True)
    return state_dir


def get_local_state_path(key: str, project_root: Path) -> Path:
    """Get the local state file path for a specific component state key.

    Args:
        key: The defs state key for the component.
        project_root: The root directory of the project.

    Returns:
        The path to the state file for this key.
    """
    return get_local_state_dir(key, project_root) / "state"


def get_code_server_metadata_key(key: str) -> str:
    """Returns a key for storing defs state in the code server reconstruction metadata.
    Avoids using the original key directly to avoid potential collisions.
    """
    return f"defs-state-[{key}]"



from typing import Annotated

from dagster_shared.record import record
from dagster_shared.serdes.objects.models.defs_state_info import DefsStateManagementType
from pydantic import Field

from dagster.components.resolved.model import Model, Resolver


class DefsStateConfigArgs(Model):
    key: str | None = Field(
        default=None, description="The key for the state. This must be unique per deployment."
    )
    management_type: DefsStateManagementType = Field(
        description="The storage type for state required for loading this object's definitions."
        "  - `LOCAL_FILESYSTEM`: State is stored on the local filesystem. `dg utils refresh-defs-state` must be executed while building the deployed container image in order for state to be accessible."
        "  - `VERSIONED_STATE_STORAGE`: State is stored in your configured `defs_state_storage`. `dg utils refresh-defs-state` may be executed at any time to refresh the state."
        "  - `LEGACY_CODE_SERVER_SNAPSHOTS`: State is stored in memory in the code server. State is always automatically refreshed when the code server is loaded.",
        examples=[
            DefsStateManagementType.LOCAL_FILESYSTEM.value,
            DefsStateManagementType.VERSIONED_STATE_STORAGE.value,
            DefsStateManagementType.LEGACY_CODE_SERVER_SNAPSHOTS.value,
        ],
    )
    refresh_if_dev: bool = Field(
        default=True,
        description="Whether to automatically refresh defs state when using `dagster dev` or the `dg` cli.",
    )

    @classmethod
    def local_filesystem(cls) -> "DefsStateConfigArgs":
        return cls(management_type=DefsStateManagementType.LOCAL_FILESYSTEM)

    @classmethod
    def versioned_state_storage(cls) -> "DefsStateConfigArgs":
        return cls(management_type=DefsStateManagementType.VERSIONED_STATE_STORAGE)

    @classmethod
    def legacy_code_server_snapshots(cls) -> "DefsStateConfigArgs":
        return cls(management_type=DefsStateManagementType.LEGACY_CODE_SERVER_SNAPSHOTS)


@record
class DefsStateConfig:
    key: str
    management_type: DefsStateManagementType
    refresh_if_dev: bool

    @classmethod
    def from_args(cls, args: DefsStateConfigArgs, default_key: str) -> "DefsStateConfig":
        return cls(
            key=args.key or default_key,
            management_type=args.management_type,
            refresh_if_dev=args.refresh_if_dev,
        )


ResolvedDefsStateConfig = Annotated[
    DefsStateConfigArgs,
    Resolver.default(
        description="Configuration for determining how state is stored and persisted for this component."
    ),
]



import json
from collections.abc import Iterable, Sequence, Set
from typing import NamedTuple

import dagster._check as check
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
from dagster._core.errors import DagsterInvalidDeserializationVersionError
from dagster._serdes import whitelist_for_serdes


@whitelist_for_serdes
class DefaultPartitionsSubset(
    PartitionsSubset,
    NamedTuple("_DefaultPartitionsSubset", [("subset", Set[str])]),
):
    # Every time we change the serialization format, we should increment the version number.
    # This will ensure that we can gracefully degrade when deserializing old data.
    SERIALIZATION_VERSION = 1

    def __new__(
        cls,
        subset: Set[str] | None = None,
    ):
        check.opt_set_param(subset, "subset")
        return super().__new__(cls, subset or set())

    @property
    def is_empty(self) -> bool:
        return len(self.subset) == 0

    def get_partition_keys_not_in_subset(
        self, partitions_def: PartitionsDefinition
    ) -> Iterable[str]:
        return [key for key in partitions_def.get_partition_keys() if key not in self.subset]

    def get_partition_keys(self) -> Iterable[str]:
        return self.subset

    def __sub__(self, other: "PartitionsSubset") -> "PartitionsSubset":
        if not isinstance(other, DefaultPartitionsSubset):
            return super().__sub__(other)

        if self is other:
            return self.empty_subset()
        if other.is_empty:
            return self

        return DefaultPartitionsSubset(self.subset - other.subset)

    def __or__(self, other: "PartitionsSubset") -> "PartitionsSubset":
        if not isinstance(other, DefaultPartitionsSubset):
            return super().__or__(other)

        if self is other or other.is_empty:
            return self

        if self.is_empty:
            return other

        return DefaultPartitionsSubset(self.subset | other.subset)

    def __and__(self, other: "PartitionsSubset") -> "PartitionsSubset":
        if not isinstance(other, DefaultPartitionsSubset):
            return super().__and__(other)

        if self is other:
            return self
        if other.is_empty:
            return other
        if self.is_empty:
            return self

        return DefaultPartitionsSubset(self.subset & other.subset)

    def get_ranges_for_keys(self, partition_keys: Sequence[str]) -> Sequence[PartitionKeyRange]:
        cur_range_start = None
        cur_range_end = None
        result = []
        for partition_key in partition_keys:
            if partition_key in self.subset:
                if cur_range_start is None:
                    cur_range_start = partition_key
                cur_range_end = partition_key
            else:
                if cur_range_start is not None and cur_range_end is not None:
                    result.append(PartitionKeyRange(cur_range_start, cur_range_end))
                cur_range_start = cur_range_end = None

        if cur_range_start is not None and cur_range_end is not None:
            result.append(PartitionKeyRange(cur_range_start, cur_range_end))
        return result

    def get_partition_key_ranges(
        self, partitions_def: PartitionsDefinition
    ) -> Sequence[PartitionKeyRange]:
        from dagster._core.definitions.partitions.definition.multi import MultiPartitionsDefinition

        if isinstance(partitions_def, MultiPartitionsDefinition):
            # For multi-partitions, we construct the ranges by holding one dimension constant
            # and constructing the range for the other dimension
            primary_dimension = partitions_def.primary_dimension
            secondary_dimension = partitions_def.secondary_dimension

            primary_keys_in_subset = set()
            secondary_keys_in_subset = set()
            for partition_key in self.subset:
                primary_keys_in_subset.add(
                    partitions_def.get_partition_key_from_str(partition_key).keys_by_dimension[
                        primary_dimension.name
                    ]
                )
                secondary_keys_in_subset.add(
                    partitions_def.get_partition_key_from_str(partition_key).keys_by_dimension[
                        secondary_dimension.name
                    ]
                )

            # for efficiency, group the keys by whichever dimension has fewer distinct keys
            grouping_dimension = (
                primary_dimension
                if len(primary_keys_in_subset) <= len(secondary_keys_in_subset)
                else secondary_dimension
            )
            grouping_keys = (
                primary_keys_in_subset
                if grouping_dimension == primary_dimension
                else secondary_keys_in_subset
            )

            results = []
            for grouping_key in grouping_keys:
                keys = partitions_def.get_multipartition_keys_with_dimension_value(
                    dimension_name=grouping_dimension.name,
                    dimension_partition_key=grouping_key,
                )
                results.extend(self.get_ranges_for_keys(keys))
            return results

        else:
            partition_keys = partitions_def.get_partition_keys()

            return self.get_ranges_for_keys(partition_keys)

    def with_partition_keys(self, partition_keys: Iterable[str]) -> "DefaultPartitionsSubset":
        return DefaultPartitionsSubset(
            self.subset | set(partition_keys),
        )

    def serialize(self) -> str:
        # Serialize version number, so attempting to deserialize old versions can be handled gracefully.
        # Any time the serialization format changes, we should increment the version number.
        return json.dumps(
            {
                "version": self.SERIALIZATION_VERSION,
                # sort to ensure that equivalent partition subsets have identical serialized forms
                "subset": sorted(list(self.subset)),
            }
        )

    @classmethod
    def from_serialized(
        cls, partitions_def: PartitionsDefinition, serialized: str
    ) -> "PartitionsSubset":
        # Check the version number, so only valid versions can be deserialized.
        data = json.loads(serialized)

        if isinstance(data, list):
            # backwards compatibility
            return cls(subset=set(data))
        else:
            if data.get("version") != cls.SERIALIZATION_VERSION:
                raise DagsterInvalidDeserializationVersionError(
                    f"Attempted to deserialize partition subset with version {data.get('version')},"
                    f" but only version {cls.SERIALIZATION_VERSION} is supported."
                )
            return cls(subset=set(data.get("subset")))

    @classmethod
    def can_deserialize(
        cls,
        partitions_def: PartitionsDefinition,
        serialized: str,
        serialized_partitions_def_unique_id: str | None,
        serialized_partitions_def_class_name: str | None,
    ) -> bool:
        if serialized_partitions_def_class_name is not None:
            return serialized_partitions_def_class_name == partitions_def.__class__.__name__

        data = json.loads(serialized)
        return isinstance(data, list) or (
            data.get("subset") is not None and data.get("version") == cls.SERIALIZATION_VERSION
        )

    def __eq__(self, other: object) -> bool:
        return isinstance(other, DefaultPartitionsSubset) and self.subset == other.subset

    __hash__ = None  # pyright: ignore[reportAssignmentType]

    def __len__(self) -> int:
        return len(self.subset)

    def __contains__(self, value) -> bool:
        return value in self.subset

    def __repr__(self) -> str:
        return f"DefaultPartitionsSubset(subset={self.subset})"

    @classmethod
    def create_empty_subset(
        cls, partitions_def: PartitionsDefinition | None = None
    ) -> "DefaultPartitionsSubset":
        return cls()

    def empty_subset(
        self,
    ) -> "DefaultPartitionsSubset":
        return DefaultPartitionsSubset()



import sys
from abc import abstractmethod
from typing import TYPE_CHECKING, Optional

from dagster import check
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.storage.dagster_run import IN_PROGRESS_RUN_STATUSES
from dagster._utils.error import serializable_error_info_from_exc_info

if TYPE_CHECKING:
    from dagster._core.instance.instance import DagsterInstance
    from dagster._core.launcher.base import RunLauncher
    from dagster._core.run_coordinator.base import BaseWorkspaceRequestContext, RunCoordinator
    from dagster._core.storage.dagster_run import DagsterRun


class RunLauncherMethods:
    """Mixin class containing run launcher-related functionality for DagsterInstance.

    This class provides methods for run submission, launching, and resumption functionality.
    All methods are implemented as instance methods that DagsterInstance inherits.
    """

    # Abstract methods that DagsterInstance provides
    @abstractmethod
    def get_run_by_id(self, run_id: str) -> Optional["DagsterRun"]: ...

    @property
    @abstractmethod
    def run_coordinator(self) -> "RunCoordinator": ...

    @property
    @abstractmethod
    def run_launcher(self) -> "RunLauncher": ...

    @property
    @abstractmethod
    def run_monitoring_enabled(self) -> bool: ...

    @property
    @abstractmethod
    def run_monitoring_max_resume_run_attempts(self) -> int: ...

    # These methods are provided by EventMethods mixin
    # (no abstract declarations needed since EventMethods implements them)

    def submit_run(self, run_id: str, workspace: "BaseWorkspaceRequestContext") -> "DagsterRun":
        """Submit a pipeline run to the coordinator.

        This method delegates to the ``RunCoordinator``, configured on the instance, and will
        call its implementation of ``RunCoordinator.submit_run()`` to send the run to the
        coordinator for execution. Runs should be created in the instance (e.g., by calling
        ``DagsterInstance.create_run()``) *before* this method is called, and
        should be in the ``PipelineRunStatus.NOT_STARTED`` state. They also must have a non-null
        ExternalPipelineOrigin.

        Args:
            run_id (str): The id of the run.
        """
        from dagster._core.run_coordinator import SubmitRunContext

        run = self.get_run_by_id(run_id)
        if run is None:
            raise DagsterInvariantViolationError(
                f"Could not load run {run_id} that was passed to submit_run"
            )

        check.not_none(
            run.remote_job_origin,
            "External pipeline origin must be set for submitted runs",
        )
        check.not_none(
            run.job_code_origin,
            "Python origin must be set for submitted runs",
        )

        try:
            submitted_run = self.run_coordinator.submit_run(
                SubmitRunContext(run, workspace=workspace)
            )
        except:
            from dagster._core.events import EngineEventData

            error = serializable_error_info_from_exc_info(sys.exc_info())
            self.report_engine_event(  # type: ignore[attr-defined]
                error.message,
                run,
                EngineEventData.engine_error(error),
            )
            self.report_run_failed(run)  # type: ignore[attr-defined]
            raise

        return submitted_run

    def launch_run(self, run_id: str, workspace: "BaseWorkspaceRequestContext") -> "DagsterRun":
        """Launch a pipeline run.

        This method is typically called using `instance.submit_run` rather than being invoked
        directly. This method delegates to the ``RunLauncher``, if any, configured on the instance,
        and will call its implementation of ``RunLauncher.launch_run()`` to begin the execution of
        the specified run. Runs should be created in the instance (e.g., by calling
        ``DagsterInstance.create_run()``) *before* this method is called, and should be in the
        ``PipelineRunStatus.NOT_STARTED`` state.

        Args:
            run_id (str): The id of the run the launch.
        """
        from dagster._core.events import DagsterEvent, DagsterEventType, EngineEventData
        from dagster._core.launcher import LaunchRunContext

        run = self.get_run_by_id(run_id)
        if run is None:
            raise DagsterInvariantViolationError(
                f"Could not load run {run_id} that was passed to launch_run"
            )

        launch_started_event = DagsterEvent(
            event_type_value=DagsterEventType.PIPELINE_STARTING.value,
            job_name=run.job_name,
        )
        self.report_dagster_event(launch_started_event, run_id=run.run_id)  # type: ignore[attr-defined]

        run = self.get_run_by_id(run_id)
        if run is None:
            check.failed(f"Failed to reload run {run_id}")

        # At this point run cannot be None due to check.failed above
        assert run is not None
        try:
            self.run_launcher.launch_run(LaunchRunContext(dagster_run=run, workspace=workspace))
        except:
            error = serializable_error_info_from_exc_info(sys.exc_info())
            self.report_engine_event(  # type: ignore[attr-defined]  # type: ignore[attr-defined]
                error.message,
                run,
                EngineEventData.engine_error(error),
            )
            self.report_run_failed(run)  # type: ignore[attr-defined]
            raise

        return run

    def resume_run(
        self,
        run_id: str,
        workspace: "BaseWorkspaceRequestContext",
        attempt_number: int,
    ) -> "DagsterRun":
        """Resume a pipeline run.

        This method should be called on runs which have already been launched, but whose run workers
        have died.

        Args:
            run_id (str): The id of the run the launch.
        """
        from dagster._core.events import EngineEventData
        from dagster._core.launcher import ResumeRunContext
        from dagster._daemon.monitoring import RESUME_RUN_LOG_MESSAGE

        run = self.get_run_by_id(run_id)
        if run is None:
            raise DagsterInvariantViolationError(
                f"Could not load run {run_id} that was passed to resume_run"
            )
        if run.status not in IN_PROGRESS_RUN_STATUSES:
            raise DagsterInvariantViolationError(
                f"Run {run_id} is not in a state that can be resumed"
            )

        self.report_engine_event(  # type: ignore[attr-defined]
            RESUME_RUN_LOG_MESSAGE,
            run,
        )

        try:
            self.run_launcher.resume_run(
                ResumeRunContext(
                    dagster_run=run,
                    workspace=workspace,
                    resume_attempt_number=attempt_number,
                )
            )
        except:
            error = serializable_error_info_from_exc_info(sys.exc_info())
            self.report_engine_event(  # type: ignore[attr-defined]  # type: ignore[attr-defined]
                error.message,
                run,
                EngineEventData.engine_error(error),
            )
            self.report_run_failed(run)  # type: ignore[attr-defined]
            raise

        return run

    def count_resume_run_attempts(self, run_id: str) -> int:
        """Count resume run attempts."""
        from typing import cast

        from dagster._daemon.monitoring import count_resume_run_attempts

        # Cast is safe since this mixin is only used by DagsterInstance
        return count_resume_run_attempts(cast("DagsterInstance", self), run_id)

    def run_will_resume(self, run_id: str) -> bool:
        """Check if run will resume."""
        if not self.run_monitoring_enabled:
            return False
        return self.count_resume_run_attempts(run_id) < self.run_monitoring_max_resume_run_attempts



import importlib
import inspect
import os
import sys
import uuid
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from pathlib import Path
from types import ModuleType
from typing import cast

from dagster_shared.record import IHaveNew, LegacyNamedTupleMixin, record, record_custom
from dagster_shared.seven import get_import_error_message, import_module_from_path
from dagster_shared.utils.hash import hash_collection

import dagster._check as check
from dagster._core.errors import DagsterImportError, DagsterInvariantViolationError
from dagster._serdes import whitelist_for_serdes
from dagster._utils import alter_sys_path


class CodePointer(ABC):
    @abstractmethod
    def load_target(self) -> object:
        pass

    @abstractmethod
    def describe(self) -> str:
        pass

    @staticmethod
    def from_module(
        module_name: str, definition: str, working_directory: str | None
    ) -> "ModuleCodePointer":
        return ModuleCodePointer(module_name, definition, working_directory)

    @staticmethod
    def from_python_package(
        module_name: str, attribute: str, working_directory: str | None
    ) -> "PackageCodePointer":
        return PackageCodePointer(module_name, attribute, working_directory)

    @staticmethod
    def from_python_file(
        python_file: str, definition: str, working_directory: str | None
    ) -> "FileCodePointer":
        return FileCodePointer(
            python_file=python_file, fn_name=definition, working_directory=working_directory
        )


def rebase_file(relative_path_in_file: str, file_path_resides_in: str) -> str:
    """In config files, you often put file paths that are meant to be relative
    to the location of that config file. This does that calculation.
    """
    check.str_param(relative_path_in_file, "relative_path_in_file")
    check.str_param(file_path_resides_in, "file_path_resides_in")
    return os.path.join(
        os.path.dirname(os.path.abspath(file_path_resides_in)), relative_path_in_file
    )


def load_python_file(
    python_file: str | Path,
    working_directory: str | None,
    add_uuid_suffix: bool = False,
) -> ModuleType:
    """Takes a path to a python file and returns a loaded module."""
    check.inst_param(python_file, "python_file", (str, Path))
    check.opt_str_param(working_directory, "working_directory")

    # First verify that the file exists
    os.stat(python_file)

    module_name = os.path.splitext(os.path.basename(python_file))[0]
    if add_uuid_suffix:
        module_name = f"{module_name}_{uuid.uuid4().hex}"

    # Use the passed in working directory for local imports (sys.path[0] isn't
    # consistently set in the different entry points that Dagster uses to import code)
    script_path = sys.path[0]
    try:
        with alter_sys_path(
            to_add=([working_directory] if working_directory else []), to_remove=[script_path]
        ):
            return import_module_from_path(module_name, python_file)
    except ImportError as ie:
        python_file = os.path.abspath(os.path.expanduser(python_file))

        msg = get_import_error_message(ie)
        if msg == "attempted relative import with no known parent package":
            raise DagsterImportError(
                f"Encountered ImportError: `{msg}` while importing module {module_name} from "
                f"file {python_file}. Consider using the module-based options `-m` for "
                "CLI-based targets or the `python_module` workspace target."
            ) from ie

        if working_directory:
            abs_working_directory = os.path.abspath(os.path.expanduser(working_directory))
            raise DagsterImportError(
                f"Encountered ImportError: `{msg}` while importing module {module_name} from "
                f"file {python_file}. Local modules were resolved using the working "
                f"directory `{abs_working_directory}`. If another working directory should be "
                "used, please explicitly specify the appropriate path using the `-d` or "
                "`--working-directory` for CLI based targets or the `working_directory` "
                "configuration option for `python_file`-based workspace targets. "
            ) from ie
        else:
            raise DagsterImportError(
                f"Encountered ImportError: `{msg}` while importing module {module_name} from file"
                f" {python_file}. If relying on the working directory to resolve modules, please "
                "explicitly specify the appropriate path using the `-d` or "
                "`--working-directory` for CLI based targets or the `working_directory` "
                "configuration option for `python_file`-based workspace targets. "
            ) from ie


def load_python_module(
    module_name: str,
    working_directory: str | None,
    remove_from_path_fn: Callable[[], Sequence[str]] | None = None,
) -> ModuleType:
    check.str_param(module_name, "module_name")
    check.opt_str_param(working_directory, "working_directory")
    check.opt_callable_param(remove_from_path_fn, "remove_from_path_fn")

    # Use the passed in working directory for local imports (sys.path[0] isn't
    # consistently set in the different entry points that Dagster uses to import code)
    remove_paths: list[str] = (
        list(remove_from_path_fn()) if remove_from_path_fn else []
    )  # hook for tests
    remove_paths.insert(0, sys.path[0])  # remove the script path
    with alter_sys_path(
        to_add=([working_directory] if working_directory else []), to_remove=remove_paths
    ):
        try:
            return importlib.import_module(module_name)
        except ImportError as ie:
            msg = get_import_error_message(ie)
            if working_directory:
                abs_working_directory = os.path.abspath(os.path.expanduser(working_directory))
                raise DagsterImportError(
                    f"Encountered ImportError: `{msg}` while importing module {module_name}. "
                    "Local modules were resolved using the working "
                    f"directory `{abs_working_directory}`. If another working directory should be "
                    "used, please explicitly specify the appropriate path using the `-d` or "
                    "`--working-directory` for CLI based targets or the `working_directory` "
                    "configuration option for workspace targets. "
                ) from ie
            else:
                raise DagsterImportError(
                    f"Encountered ImportError: `{msg}` while importing module {module_name}. "
                    "If relying on the working directory to resolve modules, please "
                    "explicitly specify the appropriate path using the `-d` or "
                    "`--working-directory` for CLI based targets or the `working_directory` "
                    "configuration option for workspace targets. "
                ) from ie


@whitelist_for_serdes
@record(kw_only=False)
class FileCodePointer(CodePointer, LegacyNamedTupleMixin):
    python_file: str
    fn_name: str
    working_directory: str | None = None

    def load_target(self) -> object:
        module = load_python_file(self.python_file, self.working_directory)
        return _load_target_from_module(
            module, self.fn_name, f"at module scope in file {self.python_file}."
        )

    def describe(self) -> str:
        if self.working_directory:
            return f"{self.python_file}::{self.fn_name} -- [dir {self.working_directory}]"
        else:
            return f"{self.python_file}::{self.fn_name}"


def _load_target_from_module(module: ModuleType, fn_name: str, error_suffix: str) -> object:
    from dagster._core.definitions.module_loaders.load_assets_from_modules import (
        load_assets_from_modules,
    )
    from dagster._core.workspace.autodiscovery import LOAD_ALL_ASSETS

    if fn_name == LOAD_ALL_ASSETS:
        # LOAD_ALL_ASSETS is a special symbol that's returned when, instead of loading a particular
        # attribute, we should load all the assets in the module.
        return load_assets_from_modules([module])
    else:
        if not hasattr(module, fn_name):
            raise DagsterInvariantViolationError(f"{fn_name} not found {error_suffix}")

        return getattr(module, fn_name)


@whitelist_for_serdes
@record(kw_only=False)
class ModuleCodePointer(CodePointer, LegacyNamedTupleMixin):
    module: str
    fn_name: str
    working_directory: str | None = None

    def load_target(self) -> object:
        module = load_python_module(self.module, self.working_directory)
        return _load_target_from_module(
            module, self.fn_name, f"in module {self.module}. dir: {dir(module)}"
        )

    def describe(self) -> str:
        return f"from {self.module} import {self.fn_name}"


@whitelist_for_serdes
@record(kw_only=False)
class PackageCodePointer(CodePointer, LegacyNamedTupleMixin):
    module: str
    attribute: str
    working_directory: str | None = None

    def load_target(self) -> object:
        module = load_python_module(self.module, self.working_directory)
        return _load_target_from_module(
            module, self.attribute, f"in module {self.module}. dir: {dir(module)}"
        )

    def describe(self) -> str:
        return f"from {self.module} import {self.attribute}"


def get_python_file_from_target(target: object) -> str | None:
    module = inspect.getmodule(target)
    python_file = getattr(module, "__file__", None)

    if not python_file:
        return None

    return os.path.abspath(python_file)


@whitelist_for_serdes
@record_custom(checked=False)
class CustomPointer(CodePointer, IHaveNew, LegacyNamedTupleMixin):
    reconstructor_pointer: ModuleCodePointer
    reconstructable_args: Sequence[object]
    reconstructable_kwargs: Sequence[Sequence]

    def __new__(
        cls,
        reconstructor_pointer: ModuleCodePointer,
        reconstructable_args: Sequence[object],
        reconstructable_kwargs: Sequence[Sequence],
    ):
        check.inst_param(reconstructor_pointer, "reconstructor_pointer", ModuleCodePointer)
        # These are lists rather than tuples to circumvent the tuple serdes machinery -- since these
        # are user-provided, they aren't whitelisted for serdes.
        check.sequence_param(reconstructable_args, "reconstructable_args")
        check.sequence_param(reconstructable_kwargs, "reconstructable_kwargs")
        for reconstructable_kwarg in reconstructable_kwargs:
            check.list_param(reconstructable_kwarg, "reconstructable_kwarg")
            check.invariant(isinstance(reconstructable_kwarg[0], str), "Bad kwarg key")
            check.invariant(
                len(reconstructable_kwarg) == 2,
                f"Bad kwarg of length {len(reconstructable_kwarg)}, should be 2",
            )

        return super().__new__(
            cls,
            reconstructor_pointer=reconstructor_pointer,
            reconstructable_args=reconstructable_args,
            reconstructable_kwargs=reconstructable_kwargs,
        )

    def load_target(self) -> object:
        reconstructor = cast("Callable", self.reconstructor_pointer.load_target())

        return reconstructor(
            *self.reconstructable_args, **{key: value for key, value in self.reconstructable_kwargs}
        )

    def describe(self) -> str:
        return f"reconstructable using {self.reconstructor_pointer.module}.{self.reconstructor_pointer.fn_name}"

    # Allow this to be hashed for use in `lru_cache`. This is needed because:
    # - `ReconstructableJob` uses `lru_cache`
    # - `ReconstructableJob` has a `ReconstructableRepository` attribute
    # - `ReconstructableRepository` has a `CodePointer` attribute
    # - `CustomCodePointer` has collection attributes that are unhashable by default
    def __hash__(self) -> int:
        if not hasattr(self, "_hash"):
            self._hash = hash_collection(
                self,  # type: ignore
            )
        return self._hash


@whitelist_for_serdes
@record
class AutoloadDefsModuleCodePointer(CodePointer):
    module: str
    working_directory: str | None

    def load_target(self) -> object:
        from dagster.components.core.load_defs import load_defs

        module = load_python_module(self.module, self.working_directory)
        return load_defs(module)

    def describe(self) -> str:
        return f"autoload from {self.module}"



import functools
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from typing import TYPE_CHECKING, Annotated, Any, Generic, TypeAlias, TypeVar, cast

from pydantic import BaseModel

from dagster import _check as check
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckSpec
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec
from dagster.components.resolved.context import ResolutionContext
from dagster.components.resolved.core_models import (
    AssetAttributesModel,
    resolve_asset_spec_update_kwargs_to_mapping,
)
from dagster.components.resolved.model import Resolver

if TYPE_CHECKING:
    from dagster.components.component.component import Component


TRANSLATOR_MERGE_ATTRIBUTES = {"metadata", "tags"}


@dataclass
class TranslatorResolvingInfo:
    asset_attributes: str | BaseModel
    resolution_context: ResolutionContext
    model_key: str = "asset_attributes"

    def _resolve_asset_attributes(self, context: Mapping[str, Any]) -> AssetSpec | BaseModel:
        """Resolves the user-specified asset attributes into an AssetAttributesModel, or an AssetSpec
        if the UDF returns one.
        """
        if not isinstance(self.asset_attributes, str):
            return self.asset_attributes

        resolved_asset_attributes = (
            self.resolution_context.at_path(self.model_key)
            .with_scope(**context)
            .resolve_value(self.asset_attributes)
        )

        if isinstance(resolved_asset_attributes, AssetSpec):
            return resolved_asset_attributes
        elif isinstance(resolved_asset_attributes, AssetAttributesModel):
            return resolved_asset_attributes
        elif isinstance(resolved_asset_attributes, dict):
            return AssetAttributesModel(**(resolved_asset_attributes))
        else:
            check.failed(
                f"Unexpected return value for asset_attributes UDF: {type(resolved_asset_attributes)}"
            )

    def get_asset_spec(self, base_spec: AssetSpec, context: Mapping[str, Any]) -> AssetSpec:
        """Returns an AssetSpec that combines the base spec with attributes resolved using the provided context.

        Usage:

        ```python
        class WrappedDagsterXTranslator(DagsterXTranslator):
            def __init__(self, *, base_translator, resolving_info: TranslatorResolvingInfo):
                self.base_translator = base_translator
                self.resolving_info = resolving_info

            def get_asset_spec(self, base_spec: AssetSpec, x_params: Any) -> AssetSpec:
                return self.resolving_info.get_asset_spec(
                    base_spec, {"x_params": x_params}
                )

        ```
        """
        resolved_asset_attributes = self._resolve_asset_attributes(context)
        if isinstance(resolved_asset_attributes, AssetSpec):
            return resolved_asset_attributes

        resolved_attributes = dict(
            resolve_asset_spec_update_kwargs_to_mapping(
                model=resolved_asset_attributes,
                context=self.resolution_context.at_path(self.model_key).with_scope(**context),
            )
        )
        if "code_version" in resolved_attributes:
            resolved_attributes = {
                **resolved_attributes,
                "code_version": str(resolved_attributes["code_version"]),
            }

        if "key_prefix" in resolved_attributes:
            prefix = resolved_attributes.pop("key_prefix")
            if "key" in resolved_attributes:
                key = resolved_attributes.pop("key")
            else:
                key = base_spec.key
            key = key.with_prefix(prefix)
            resolved_attributes["key"] = key

        return base_spec.replace_attributes(
            **{k: v for k, v in resolved_attributes.items() if k not in TRANSLATOR_MERGE_ATTRIBUTES}
        ).merge_attributes(
            **{k: v for k, v in resolved_attributes.items() if k in TRANSLATOR_MERGE_ATTRIBUTES}
        )


T = TypeVar("T")

TranslationFn: TypeAlias = Callable[[AssetSpec, T], AssetSpec]


def _build_translation_fn(
    template_vars_for_translation_fn: Callable[[T], Mapping[str, Any]],
) -> Callable[[ResolutionContext, T], TranslationFn[T]]:
    def resolve_translation(
        context: ResolutionContext,
        model,
    ) -> TranslationFn[T]:
        info = TranslatorResolvingInfo(
            asset_attributes=model,
            resolution_context=context,
            model_key="translation",
        )
        return lambda base_asset_spec, data: info.get_asset_spec(
            base_asset_spec,
            {
                **(template_vars_for_translation_fn(data)),
                "spec": base_asset_spec,
            },
        )

    return resolve_translation


class TranslationFnResolver(Resolver, Generic[T]):
    """Resolver which builds a TranslationFn from input AssetAttributesModel, injecting
    the provided template vars into the resolution process.

    This is useful to expose nested fields within the input data as template vars.

    Example:
    .. code-block:: python

        class MyApiData(BaseModel):
            name: str

        class MyApiComponent(Component, Resolvable):
            translation: Annotated[
                TranslationFn[MyApiData],
                TranslationFnResolver[MyApiData](
                    template_vars_for_translation_fn=lambda data: {"name": data.name}
                ),
            ]

    .. code-block:: yaml

        type: my_module.MyApiComponent

        attributes:
            translation:
                key: {{ name }}

    """

    def __init__(
        self,
        template_vars_for_translation_fn: Callable[[T], Mapping[str, Any]],
        model_field_type: type = AssetAttributesModel,
    ):
        super().__init__(
            _build_translation_fn(
                template_vars_for_translation_fn=template_vars_for_translation_fn
            ),
            model_field_type=str | model_field_type,
            inject_before_resolve=False,
        )


# Common case of a TranslationFn that takes a single underlying entity and passes it through
# as a template var called "data".
ResolvedTranslationFn: TypeAlias = (
    Annotated[TranslationFn[T], TranslationFnResolver[T](lambda data: {"data": data})] | None
)

T_Component = TypeVar("T_Component", bound="Component", covariant=True)
T_Translator = TypeVar("T_Translator")


class ComponentTranslator(Generic[T_Component]):
    """To support python versions < 3.10, we need to use a Protocol to tell the type system that
    these generated classes have a component property.
    """

    def __init__(self, component: T_Component, *args, **kwargs):
        self._component = component
        super().__init__(*args, **kwargs)

    @property
    def component(self) -> T_Component:
        return self._component


def create_component_translator_cls(
    base_component_cls: type[T_Component], base_translator_cls: type[T_Translator]
) -> type[T_Translator]:
    class _GeneratedComponentTranslator(base_translator_cls, ComponentTranslator[T_Component]):  # type: ignore
        def _shim_method(self, method_name: str) -> Callable:
            component_base_method = getattr(base_component_cls, method_name)
            component_instance_method = getattr(self._component.__class__, method_name)
            if component_base_method is not component_instance_method:
                # if the user has overridden the method, we invoke the instance method
                return functools.partial(component_instance_method, self._component)
            else:
                # we never invoke the component_base_method directly. instead, if the user
                # has not overridden the method, we invoke the original translator class method
                return getattr(super(), method_name)

        def get_asset_spec(self, *args, **kwargs) -> AssetSpec:
            return self._shim_method("get_asset_spec")(*args, **kwargs)

        def get_asset_check_spec(self, *args, **kwargs) -> AssetCheckSpec | None:
            return self._shim_method("get_asset_check_spec")(*args, **kwargs)

    return cast("type[T_Translator]", _GeneratedComponentTranslator)



from datetime import datetime
from typing import TYPE_CHECKING, NamedTuple, Optional

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.definitions.partitions.definition import PartitionsDefinition
from dagster._core.definitions.partitions.mapping.partition_mapping import (
    PartitionMapping,
    UpstreamPartitionsResult,
)
from dagster._core.definitions.partitions.subset.default import DefaultPartitionsSubset
from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.instance import DynamicPartitionsStore


@whitelist_for_serdes
@public
class IdentityPartitionMapping(PartitionMapping, NamedTuple("_IdentityPartitionMapping", [])):
    """Expects that the upstream and downstream assets are partitioned in the same way, and maps
    partitions in the downstream asset to the same partition key in the upstream asset.

    .. code-block:: python

        import dagster as dg

        daily_partitions_def = dg.DailyPartitionsDefinition(start_date="2025-01-01")

        alternating_daily_partitions_def = dg.TimeWindowPartitionsDefinition(
            start="2025-01-01",
            fmt="%Y-%m-%d",
            cron_schedule="0 0 */2 * *",
        )


        @dg.asset(partitions_def=daily_partitions_def)
        def asset_upstream(context: dg.AssetExecutionContext): ...


        # Downstream asset will map to the upstream when it is the same day
        @dg.asset(
            partitions_def=alternating_daily_partitions_def,
            deps=[
                dg.AssetDep(
                    asset=asset_upstream, partition_mapping=dg.IdentityPartitionMapping()
                )
            ],
        )
        def asset_downstream(context: dg.AssetExecutionContext): ...

    """

    def validate_partition_mapping(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition | None,
    ):
        if type(upstream_partitions_def) != type(downstream_partitions_def):
            raise DagsterInvalidDefinitionError(
                "Upstream and downstream partitions definitions must match, or a different partition mapping must be provided. "
                f"Got upstream definition {type(upstream_partitions_def)} and downstream definition {type(downstream_partitions_def)}",
            )

    def get_upstream_mapped_partitions_result_for_partitions(
        self,
        downstream_partitions_subset: PartitionsSubset | None,
        downstream_partitions_def: PartitionsDefinition | None,
        upstream_partitions_def: PartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> UpstreamPartitionsResult:
        with partition_loading_context(current_time, dynamic_partitions_store):
            if downstream_partitions_subset is None:
                check.failed("downstream asset is not partitioned")

            if downstream_partitions_def == upstream_partitions_def:
                return UpstreamPartitionsResult(
                    partitions_subset=downstream_partitions_subset,
                    required_but_nonexistent_subset=upstream_partitions_def.empty_subset(),
                )

            # must list out the keys before combining them since they might be from
            # different asset keys
            upstream_partition_keys = set(upstream_partitions_def.get_partition_keys())
            downstream_partition_keys = set(downstream_partitions_subset.get_partition_keys())

            return UpstreamPartitionsResult(
                partitions_subset=upstream_partitions_def.subset_with_partition_keys(
                    list(upstream_partition_keys & downstream_partition_keys)
                ),
                required_but_nonexistent_subset=DefaultPartitionsSubset(
                    downstream_partition_keys - upstream_partition_keys,
                ),
            )

    def get_downstream_partitions_for_partitions(
        self,
        upstream_partitions_subset: PartitionsSubset,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> PartitionsSubset:
        with partition_loading_context(current_time, dynamic_partitions_store):
            if upstream_partitions_subset is None:
                check.failed("upstream asset is not partitioned")

            if upstream_partitions_def == downstream_partitions_def:
                return upstream_partitions_subset

            upstream_partition_keys = set(upstream_partitions_subset.get_partition_keys())
            downstream_partition_keys = set(downstream_partitions_def.get_partition_keys())

            return downstream_partitions_def.empty_subset().with_partition_keys(
                list(downstream_partition_keys & upstream_partition_keys)
            )

    @property
    def description(self) -> str:
        return (
            "Assumes upstream and downstream assets share the same partitions definition. "
            "Maps each partition in the downstream asset to the same partition in the upstream asset."
        )



import sys
import threading
from abc import abstractmethod
from collections import defaultdict
from collections.abc import Callable, Mapping, Sequence
from contextlib import AbstractContextManager
from functools import cached_property, partial
from typing import TYPE_CHECKING, Any, Optional, Union, cast

from dagster_shared.libraries import DagsterLibraryRegistry
from dagster_shared.serdes.objects.models.defs_state_info import DefsStateInfo

import dagster._check as check
from dagster._check import checked
from dagster._core.code_pointer import CodePointer
from dagster._core.definitions.assets.job.asset_job import IMPLICIT_ASSET_JOB_NAME
from dagster._core.definitions.reconstruct import ReconstructableJob, ReconstructableRepository
from dagster._core.definitions.repository_definition import RepositoryDefinition
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.definitions.timestamp import TimestampWithTimezone
from dagster._core.errors import DagsterInvalidSubsetError, DagsterUserCodeProcessError
from dagster._core.execution.api import create_execution_plan
from dagster._core.execution.plan.state import KnownExecutionState
from dagster._core.instance import DagsterInstance
from dagster._core.origin import RepositoryPythonOrigin
from dagster._core.remote_origin import (
    CodeLocationOrigin,
    GrpcServerCodeLocationOrigin,
    InProcessCodeLocationOrigin,
)
from dagster._core.remote_representation.external import (
    RemoteExecutionPlan,
    RemoteJob,
    RemoteRepository,
)
from dagster._core.remote_representation.external_data import (
    PartitionNamesSnap,
    RemoteJobSubsetResult,
    RepositorySnap,
    ScheduleExecutionErrorSnap,
    SensorExecutionErrorSnap,
)
from dagster._core.remote_representation.grpc_server_registry import GrpcServerRegistry
from dagster._core.remote_representation.handle import JobHandle, RepositoryHandle
from dagster._core.snap.execution_plan_snapshot import snapshot_from_execution_plan
from dagster._grpc.impl import (
    get_external_schedule_execution,
    get_external_sensor_execution,
    get_notebook_data,
    get_partition_config,
    get_partition_names,
    get_partition_set_execution_param_data,
    get_partition_tags,
)
from dagster._grpc.types import GetCurrentImageResult, GetCurrentRunsResult
from dagster._record import copy
from dagster._serdes import deserialize_value
from dagster._utils.merger import merge_dicts

if TYPE_CHECKING:
    from dagster._core.definitions.schedule_definition import ScheduleExecutionData
    from dagster._core.definitions.sensor_definition import SensorExecutionData
    from dagster._core.remote_representation.external_data import (
        JobDataSnap,
        JobRefSnap,
        PartitionConfigSnap,
        PartitionExecutionErrorSnap,
        PartitionSetExecutionParamSnap,
        PartitionTagsSnap,
    )


class CodeLocation(AbstractContextManager):
    """A CodeLocation represents a target containing user code which has a set of Dagster
    definition objects. A given location will contain some number of uniquely named
    RepositoryDefinitions, which therein contains job, op, and other definitions.

    Dagster tools are typically "host" processes, meaning they load a CodeLocation and
    communicate with it over an IPC/RPC layer. Currently this IPC layer is implemented by
    invoking the dagster CLI in a target python interpreter (e.g. a virtual environment) in either
      a) the current node
      b) a container

    In the near future, we may also make this communication channel able over an RPC layer, in
    which case the information needed to load a CodeLocation will be a url that abides by
    some RPC contract.

    We also allow for InProcessCodeLocation which actually loads the user-defined artifacts
    into process with the host tool. This is mostly for test scenarios.
    """

    @abstractmethod
    def get_repository(self, name: str) -> RemoteRepository:
        pass

    @abstractmethod
    def has_repository(self, name: str) -> bool:
        pass

    @abstractmethod
    def get_repositories(self) -> Mapping[str, RemoteRepository]:
        pass

    def get_repository_names(self) -> Sequence[str]:
        return list(self.get_repositories().keys())

    @property
    def name(self) -> str:
        return self.origin.location_name

    @abstractmethod
    def get_execution_plan(
        self,
        remote_job: RemoteJob,
        run_config: Mapping[str, object],
        step_keys_to_execute: Sequence[str] | None,
        known_state: KnownExecutionState | None,
        instance: DagsterInstance | None = None,
    ) -> RemoteExecutionPlan: ...

    @abstractmethod
    async def gen_execution_plan(
        self,
        remote_job: RemoteJob,
        run_config: Mapping[str, object],
        step_keys_to_execute: Sequence[str] | None,
        known_state: KnownExecutionState | None,
        instance: DagsterInstance | None = None,
    ) -> RemoteExecutionPlan: ...

    def _get_remote_job_from_subset_result(
        self,
        selector: JobSubsetSelector,
        subset_result: RemoteJobSubsetResult,
    ) -> RemoteJob:
        if subset_result.repository_python_origin:
            # Prefer the python origin from the result if it is set, in case the code location
            # just updated and any origin information (most frequently the image) has changed
            repo_handle = RepositoryHandle(
                repository_name=selector.repository_name,
                code_location_origin=self.origin,
                repository_python_origin=subset_result.repository_python_origin,
                display_metadata=self.get_display_metadata(),
            )
        else:
            repo_handle = RepositoryHandle.from_location(
                repository_name=selector.repository_name,
                code_location=self,
            )

        job_data_snap = subset_result.job_data_snap
        if job_data_snap is None:
            error = check.not_none(subset_result.error)
            if error.cls_name == "DagsterInvalidSubsetError":
                raise DagsterInvalidSubsetError(check.not_none(error.message))
            else:
                check.failed(
                    f"Failed to fetch subset data, success: {subset_result.success} error: {error}"
                )

        return RemoteJob(job_data_snap, repo_handle)

    def get_job(self, selector: JobSubsetSelector) -> RemoteJob:
        """Return the RemoteJob for a specific pipeline. Subclasses only
        need to implement get_subset_remote_job_result to handle the case where
        an op selection is specified, which requires access to the underlying JobDefinition
        to generate the subsetted pipeline snapshot.
        """
        if not selector.is_subset_selection:
            return self.get_repository(selector.repository_name).get_full_job(selector.job_name)

        subset_result = self._get_subset_remote_job_result(
            selector,
            lambda selector: self.get_repository(selector.repository_name).get_full_job(
                selector.job_name
            ),
        )
        return self._get_remote_job_from_subset_result(selector, subset_result)

    async def gen_subset_job(
        self, selector: JobSubsetSelector, get_full_job: Callable[[JobSubsetSelector], RemoteJob]
    ) -> RemoteJob:
        subset_result = await self._gen_subset_remote_job_result(selector, get_full_job)
        return self._get_remote_job_from_subset_result(selector, subset_result)

    @abstractmethod
    def _get_subset_remote_job_result(
        self, selector: JobSubsetSelector, get_full_job: Callable[[JobSubsetSelector], RemoteJob]
    ) -> RemoteJobSubsetResult:
        """Returns a snapshot about an RemoteJob with an op selection, which requires
        access to the underlying JobDefinition. Callsites should likely use
        `get_job` instead.
        """

    @abstractmethod
    async def _gen_subset_remote_job_result(
        self, selector: JobSubsetSelector, get_full_job: Callable[[JobSubsetSelector], RemoteJob]
    ) -> RemoteJobSubsetResult:
        """Returns a snapshot about an RemoteJob with an op selection, which requires
        access to the underlying JobDefinition. Callsites should likely use
        `gen_job` instead.
        """

    @abstractmethod
    def get_partition_config(
        self,
        repository_handle: RepositoryHandle,
        job_name: str,
        partition_name: str,
        instance: DagsterInstance,
    ) -> Union["PartitionConfigSnap", "PartitionExecutionErrorSnap"]:
        pass

    @abstractmethod
    def get_partition_tags_from_repo(
        self,
        repository_handle: RepositoryHandle,
        job_name: str,
        partition_name: str,
        instance: DagsterInstance,
    ) -> Union["PartitionTagsSnap", "PartitionExecutionErrorSnap"]:
        pass

    @abstractmethod
    def get_partition_names_from_repo(
        self,
        repository_handle: RepositoryHandle,
        job_name: str,
    ) -> Union["PartitionNamesSnap", "PartitionExecutionErrorSnap"]:
        pass

    @abstractmethod
    def get_partition_set_execution_params(
        self,
        repository_handle: RepositoryHandle,
        partition_set_name: str,
        partition_names: Sequence[str],
        instance: DagsterInstance,
    ) -> Union["PartitionSetExecutionParamSnap", "PartitionExecutionErrorSnap"]:
        pass

    @abstractmethod
    def get_schedule_execution_data(
        self,
        instance: DagsterInstance,
        repository_handle: RepositoryHandle,
        schedule_name: str,
        scheduled_execution_time: TimestampWithTimezone | None,
        log_key: Sequence[str] | None,
    ) -> "ScheduleExecutionData":
        pass

    @abstractmethod
    def get_sensor_execution_data(
        self,
        instance: DagsterInstance,
        repository_handle: RepositoryHandle,
        name: str,
        last_tick_completion_time: float | None,
        last_run_key: str | None,
        cursor: str | None,
        log_key: Sequence[str] | None,
        last_sensor_start_time: float | None,
    ) -> "SensorExecutionData":
        pass

    @abstractmethod
    def get_notebook_data(self, notebook_path: str) -> bytes:
        pass

    @property
    @abstractmethod
    def is_reload_supported(self) -> bool:
        pass

    def __del__(self):
        self.cleanup()

    def __exit__(self, _exception_type, _exception_value, _traceback):
        self.cleanup()

    def cleanup(self) -> None:
        pass

    @property
    @abstractmethod
    def origin(self) -> CodeLocationOrigin:
        pass

    def get_display_metadata(self) -> Mapping[str, str]:
        return merge_dicts(
            self.origin.get_display_metadata(),
            ({"image": self.container_image} if self.container_image else {}),
        )

    @property
    @abstractmethod
    def executable_path(self) -> str | None:
        pass

    @property
    @abstractmethod
    def container_image(self) -> str | None:
        pass

    @cached_property
    def container_context(self) -> Mapping[str, Any] | None:
        return None

    @property
    @abstractmethod
    def entry_point(self) -> Sequence[str] | None:
        pass

    @property
    @abstractmethod
    def repository_code_pointer_dict(self) -> Mapping[str, CodePointer | None]:
        pass

    def get_repository_python_origin(
        self, repository_name: str
    ) -> Optional["RepositoryPythonOrigin"]:
        code_pointer = self.repository_code_pointer_dict.get(repository_name)
        if not code_pointer:
            return None

        return RepositoryPythonOrigin(
            executable_path=self.executable_path or sys.executable,
            code_pointer=code_pointer,
            container_image=self.container_image,
            entry_point=self.entry_point,
            container_context=self.container_context,
        )

    @abstractmethod
    def get_dagster_library_versions(self) -> Mapping[str, str] | None: ...

    def get_defs_state_info(self) -> DefsStateInfo | None:
        all_infos = list(
            filter(
                None,
                [repo.repository_snap.defs_state_info for repo in self.get_repositories().values()],
            )
        )
        if len(all_infos) == 0:
            return None
        elif len(all_infos) == 1:
            return all_infos[0]
        else:
            # in theory this would be extremely rare, as having multiple
            # repositories in the same location has long been deprecated
            combined_mapping = {}
            for info in all_infos:
                combined_mapping.update(info.info_mapping)
            return DefsStateInfo(info_mapping=combined_mapping)


class InProcessCodeLocation(CodeLocation):
    def __init__(self, origin: InProcessCodeLocationOrigin, instance: DagsterInstance):
        from dagster._grpc.server import LoadedRepositories

        self._origin = check.inst_param(origin, "origin", InProcessCodeLocationOrigin)
        self._instance = instance

        loadable_target_origin = self._origin.loadable_target_origin
        self._loaded_repositories = LoadedRepositories(
            loadable_target_origin,
            entry_point=self._origin.entry_point,
            container_image=self._origin.container_image,
            container_context=self._origin.container_context,
            # for InProcessCodeLocations, we always use the latest available state versions
            defs_state_info=self._instance.defs_state_storage.get_latest_defs_state_info()
            if self._instance.defs_state_storage
            else None,
        )

        self._repository_code_pointer_dict = self._loaded_repositories.code_pointers_by_repo_name

        self._repositories: dict[str, RemoteRepository] = {}
        for (
            repo_name,
            repo_def,
        ) in self._loaded_repositories.definitions_by_name.items():
            self._repositories[repo_name] = RemoteRepository(
                RepositorySnap.from_def(repo_def),
                RepositoryHandle.from_location(repository_name=repo_name, code_location=self),
                auto_materialize_use_sensors=instance.auto_materialize_use_sensors,
            )

    @property
    def is_reload_supported(self) -> bool:
        return False

    @property
    def origin(self) -> InProcessCodeLocationOrigin:
        return self._origin

    @property
    def executable_path(self) -> str | None:
        return self._origin.loadable_target_origin.executable_path

    @property
    def container_image(self) -> str | None:
        return self._origin.container_image

    @cached_property
    def container_context(self) -> Mapping[str, Any] | None:
        return self._origin.container_context

    @property
    def entry_point(self) -> Sequence[str] | None:
        return self._origin.entry_point

    @property
    def repository_code_pointer_dict(self) -> Mapping[str, CodePointer | None]:
        return self._repository_code_pointer_dict

    def _get_reconstructable_repository(self, repository_name: str) -> ReconstructableRepository:
        return self._loaded_repositories.reconstructables_by_name[repository_name]

    def get_reconstructable_job(self, repository_name: str, name: str) -> ReconstructableJob:
        return self._get_reconstructable_repository(repository_name).get_reconstructable_job(name)

    def _get_repo_def(self, name: str) -> RepositoryDefinition:
        return self._loaded_repositories.definitions_by_name[name]

    def get_repository(self, name: str) -> RemoteRepository:
        return self._repositories[name]

    def has_repository(self, name: str) -> bool:
        return name in self._repositories

    def get_repositories(self) -> Mapping[str, RemoteRepository]:
        return self._repositories

    async def _gen_subset_remote_job_result(
        self, selector: JobSubsetSelector, get_full_job: Callable[[JobSubsetSelector], RemoteJob]
    ) -> RemoteJobSubsetResult:
        return self._get_subset_remote_job_result(selector, get_full_job)

    def _get_subset_remote_job_result(
        self, selector: JobSubsetSelector, get_full_job: Callable[[JobSubsetSelector], RemoteJob]
    ) -> RemoteJobSubsetResult:
        check.inst_param(selector, "selector", JobSubsetSelector)
        check.invariant(
            selector.location_name == self.name,
            f"PipelineSelector location_name mismatch, got {selector.location_name} expected"
            f" {self.name}",
        )

        from dagster._grpc.impl import get_external_pipeline_subset_result

        return get_external_pipeline_subset_result(
            self._get_repo_def(selector.repository_name),
            self._get_reconstructable_repository(selector.repository_name),
            selector.job_name,
            selector.op_selection,
            selector.asset_selection,
            selector.asset_check_selection,
            include_parent_snapshot=True,
        )

    async def gen_execution_plan(
        self,
        remote_job: RemoteJob,
        run_config: Mapping[str, object],
        step_keys_to_execute: Sequence[str] | None,
        known_state: KnownExecutionState | None,
        instance: DagsterInstance | None = None,
    ) -> RemoteExecutionPlan:
        return self.get_execution_plan(
            remote_job,
            run_config,
            step_keys_to_execute,
            known_state,
            instance,
        )

    def get_execution_plan(
        self,
        remote_job: RemoteJob,
        run_config: Mapping[str, object],
        step_keys_to_execute: Sequence[str] | None,
        known_state: KnownExecutionState | None,
        instance: DagsterInstance | None = None,
    ) -> RemoteExecutionPlan:
        check.inst_param(remote_job, "remote_job", RemoteJob)
        check.mapping_param(run_config, "run_config")
        step_keys_to_execute = check.opt_nullable_sequence_param(
            step_keys_to_execute, "step_keys_to_execute", of_type=str
        )
        check.opt_inst_param(known_state, "known_state", KnownExecutionState)
        check.opt_inst_param(instance, "instance", DagsterInstance)

        execution_plan = create_execution_plan(
            job=self.get_reconstructable_job(
                remote_job.repository_handle.repository_name, remote_job.name
            ).get_subset(
                op_selection=remote_job.resolved_op_selection,
                asset_selection=remote_job.asset_selection,
                asset_check_selection=remote_job.asset_check_selection,
            ),
            run_config=run_config,
            step_keys_to_execute=step_keys_to_execute,
            known_state=known_state,
            instance_ref=instance.get_ref() if instance and instance.is_persistent else None,
        )
        return RemoteExecutionPlan(
            execution_plan_snapshot=snapshot_from_execution_plan(
                execution_plan,
                remote_job.identifying_job_snapshot_id,
            )
        )

    def get_partition_config(
        self,
        repository_handle: RepositoryHandle,
        job_name: str,
        partition_name: str,
        instance: DagsterInstance,
    ) -> Union["PartitionConfigSnap", "PartitionExecutionErrorSnap"]:
        check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
        check.str_param(job_name, "job_name")
        check.str_param(partition_name, "partition_name")

        return get_partition_config(
            self._get_repo_def(repository_handle.repository_name),
            job_name=job_name,
            partition_key=partition_name,
            instance_ref=instance.get_ref(),
        )

    def get_partition_tags_from_repo(
        self,
        repository_handle: RepositoryHandle,
        job_name: str,
        partition_name: str,
        instance: DagsterInstance,
    ) -> Union["PartitionTagsSnap", "PartitionExecutionErrorSnap"]:
        check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
        check.str_param(job_name, "job_name")
        check.str_param(partition_name, "partition_name")
        check.inst_param(instance, "instance", DagsterInstance)

        return get_partition_tags(
            self._get_repo_def(repository_handle.repository_name),
            job_name=job_name,
            partition_name=partition_name,
            instance_ref=instance.get_ref(),
        )

    def get_partition_names_from_repo(
        self,
        repository_handle: RepositoryHandle,
        job_name: str,
    ) -> Union["PartitionNamesSnap", "PartitionExecutionErrorSnap"]:
        return get_partition_names(
            self._get_repo_def(repository_handle.repository_name),
            job_name=job_name,
        )

    def get_schedule_execution_data(
        self,
        instance: DagsterInstance,
        repository_handle: RepositoryHandle,
        schedule_name: str,
        scheduled_execution_time: TimestampWithTimezone | None,
        log_key: Sequence[str] | None,
    ) -> "ScheduleExecutionData":
        check.inst_param(instance, "instance", DagsterInstance)
        check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
        check.str_param(schedule_name, "schedule_name")
        check.opt_inst_param(
            scheduled_execution_time, "scheduled_execution_time", TimestampWithTimezone
        )
        check.opt_list_param(log_key, "log_key", of_type=str)

        result = get_external_schedule_execution(
            self._get_repo_def(repository_handle.repository_name),
            instance_ref=instance.get_ref(),
            schedule_name=schedule_name,
            scheduled_execution_timestamp=(
                scheduled_execution_time.timestamp if scheduled_execution_time else None
            ),
            scheduled_execution_timezone=(
                scheduled_execution_time.timezone if scheduled_execution_time else None
            ),
            log_key=log_key,
        )
        if isinstance(result, ScheduleExecutionErrorSnap):
            raise DagsterUserCodeProcessError.from_error_info(result.error)

        return result

    def get_sensor_execution_data(
        self,
        instance: DagsterInstance,
        repository_handle: RepositoryHandle,
        name: str,
        last_tick_completion_time: float | None,
        last_run_key: str | None,
        cursor: str | None,
        log_key: Sequence[str] | None,
        last_sensor_start_time: float | None,
    ) -> "SensorExecutionData":
        result = get_external_sensor_execution(
            self._get_repo_def(repository_handle.repository_name),
            self.origin,
            instance.get_ref(),
            name,
            last_tick_completion_time,
            last_run_key,
            cursor,
            log_key,
            last_sensor_start_time,
        )
        if isinstance(result, SensorExecutionErrorSnap):
            raise DagsterUserCodeProcessError.from_error_info(result.error)

        return result

    def get_partition_set_execution_params(
        self,
        repository_handle: RepositoryHandle,
        partition_set_name: str,
        partition_names: Sequence[str],
        instance: DagsterInstance,
    ) -> Union["PartitionSetExecutionParamSnap", "PartitionExecutionErrorSnap"]:
        check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
        check.str_param(partition_set_name, "partition_set_name")
        check.sequence_param(partition_names, "partition_names", of_type=str)

        return get_partition_set_execution_param_data(
            self._get_repo_def(repository_handle.repository_name),
            partition_set_name=partition_set_name,
            partition_names=partition_names,
            instance_ref=instance.get_ref(),
        )

    def get_notebook_data(self, notebook_path: str) -> bytes:
        check.str_param(notebook_path, "notebook_path")
        return get_notebook_data(notebook_path)

    def get_dagster_library_versions(self) -> Mapping[str, str]:
        return DagsterLibraryRegistry.get()


class GrpcServerCodeLocation(CodeLocation):
    def __init__(
        self,
        origin: CodeLocationOrigin,
        instance: DagsterInstance,
        host: str | None = None,
        port: int | None = None,
        socket: str | None = None,
        heartbeat: bool | None = False,
        watch_server: bool | None = True,
        grpc_server_registry: GrpcServerRegistry | None = None,
        grpc_metadata: Sequence[tuple[str, str]] | None = None,
    ):
        from dagster._api.get_server_id import sync_get_server_id
        from dagster._api.list_repositories import sync_list_repositories_grpc
        from dagster._api.snapshot_repository import sync_get_external_repositories_data_grpc
        from dagster._grpc.client import DagsterGrpcClient, client_heartbeat_thread

        self._origin = check.inst_param(origin, "origin", CodeLocationOrigin)
        self._instance = instance

        self.grpc_server_registry = check.opt_inst_param(
            grpc_server_registry, "grpc_server_registry", GrpcServerRegistry
        )

        if isinstance(self.origin, GrpcServerCodeLocationOrigin):
            self._port = self.origin.port
            self._socket = self.origin.socket
            self._host = self.origin.host
            self._use_ssl = bool(self.origin.use_ssl)
        else:
            self._port = check.opt_int_param(port, "port")
            self._socket = check.opt_str_param(socket, "socket")
            self._host = check.str_param(host, "host")
            self._use_ssl = False

        self._heartbeat_shutdown_event = None
        self._heartbeat_thread = None

        self._heartbeat = check.bool_param(heartbeat, "heartbeat")
        self._watch_server = check.bool_param(watch_server, "watch_server")

        self._server_id = None

        self._executable_path = None
        self._container_image = None
        self._container_context = None
        self._repository_code_pointer_dict = None
        self._entry_point = None

        try:
            self.client = DagsterGrpcClient(
                port=self._port,
                socket=self._socket,
                host=self._host,
                use_ssl=self._use_ssl,
                metadata=grpc_metadata,
            )

            list_repositories_response = sync_list_repositories_grpc(self.client)

            self._server_id = sync_get_server_id(self.client)
            self.repository_names = set(
                symbol.repository_name for symbol in list_repositories_response.repository_symbols
            )

            if self._heartbeat:
                self._heartbeat_shutdown_event = threading.Event()

                self._heartbeat_thread = threading.Thread(
                    target=client_heartbeat_thread,
                    args=(
                        self.client,
                        self._heartbeat_shutdown_event,
                    ),
                    name="grpc-client-heartbeat",
                    daemon=True,
                )
                self._heartbeat_thread.start()

            self._executable_path = list_repositories_response.executable_path
            self._repository_code_pointer_dict = (
                list_repositories_response.repository_code_pointer_dict
            )
            self._entry_point = list_repositories_response.entry_point
            self._dagster_library_versions = list_repositories_response.dagster_library_versions
            self._container_image = (
                list_repositories_response.container_image
                or self._reload_current_image()  # Back-compat for older gRPC servers that did not include container_image in ListRepositoriesResponse
            )

            self._container_context = list_repositories_response.container_context

            self._job_snaps_by_name = defaultdict(dict)
            self._job_snaps_by_snapshot_id = {}

            self.remote_repositories = {}

            for repo_name, (repo_data, job_snaps) in sync_get_external_repositories_data_grpc(
                self.client,
                self,
                defer_snapshots=True,
            ).items():
                self.remote_repositories[repo_name] = RemoteRepository(
                    repo_data,
                    RepositoryHandle.from_location(
                        repository_name=repo_name,
                        code_location=self,
                    ),
                    auto_materialize_use_sensors=instance.auto_materialize_use_sensors,
                    ref_to_data_fn=partial(self._job_ref_to_snap, repo_name),
                )
                for job_snap in job_snaps.values():
                    self._job_snaps_by_name[repo_name][job_snap.name] = job_snap
                    self._job_snaps_by_snapshot_id[job_snap.snapshot_id] = job_snap

        except:
            self.cleanup()
            raise

    def _job_ref_to_snap(self, repository_name: str, job_ref: "JobRefSnap") -> "JobDataSnap":
        from dagster._core.remote_representation.external_data import JobDataSnap

        # key by name to ensure that we are resilient to snapshot ID instability
        snapshot = self._job_snaps_by_name[repository_name][job_ref.name]
        parent_snapshot = (
            self._job_snaps_by_snapshot_id.get(job_ref.parent_snapshot_id)
            if job_ref.parent_snapshot_id
            else None
        )
        return JobDataSnap(
            name=job_ref.name,
            job=snapshot,
            parent_job=parent_snapshot,
            active_presets=job_ref.active_presets,
        )

    @property
    def server_id(self) -> str:
        return check.not_none(self._server_id)

    @property
    def origin(self) -> CodeLocationOrigin:
        return self._origin

    @property
    def container_image(self) -> str:
        return cast("str", self._container_image)

    @cached_property
    def container_context(self) -> Mapping[str, Any] | None:
        return self._container_context

    @property
    def repository_code_pointer_dict(self) -> Mapping[str, CodePointer | None]:
        return cast("Mapping[str, CodePointer | None]", self._repository_code_pointer_dict)

    @property
    def executable_path(self) -> str | None:
        return self._executable_path

    @property
    def entry_point(self) -> Sequence[str] | None:
        return self._entry_point

    @property
    def port(self) -> int | None:
        return self._port

    @property
    def socket(self) -> str | None:
        return self._socket

    @property
    def host(self) -> str:
        return self._host

    @property
    def use_ssl(self) -> bool:
        return self._use_ssl

    def _reload_current_image(self) -> str | None:
        return deserialize_value(
            self.client.get_current_image(),
            GetCurrentImageResult,
        ).current_image

    def get_current_runs(self) -> Sequence[str]:
        return deserialize_value(self.client.get_current_runs(), GetCurrentRunsResult).current_runs

    def cleanup(self) -> None:
        if self._heartbeat_shutdown_event:
            self._heartbeat_shutdown_event.set()
            self._heartbeat_shutdown_event = None

        if self._heartbeat_thread:
            self._heartbeat_thread.join()
            self._heartbeat_thread = None

    @property
    def is_reload_supported(self) -> bool:
        return True

    def get_repository(self, name: str) -> RemoteRepository:
        check.str_param(name, "name")
        return self.get_repositories()[name]

    def has_repository(self, name: str) -> bool:
        return name in self.get_repositories()

    def get_repositories(self) -> Mapping[str, RemoteRepository]:
        return self.remote_repositories

    def get_execution_plan(
        self,
        remote_job: RemoteJob,
        run_config: Mapping[str, Any],
        step_keys_to_execute: Sequence[str] | None,
        known_state: KnownExecutionState | None,
        instance: DagsterInstance | None = None,
    ) -> RemoteExecutionPlan:
        from dagster._api.snapshot_execution_plan import sync_get_external_execution_plan_grpc

        check.inst_param(remote_job, "remote_job", RemoteJob)
        run_config = check.mapping_param(run_config, "run_config")
        check.opt_nullable_sequence_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)
        check.opt_inst_param(known_state, "known_state", KnownExecutionState)
        check.opt_inst_param(instance, "instance", DagsterInstance)

        asset_selection = (
            frozenset(check.opt_set_param(remote_job.asset_selection, "asset_selection"))
            if remote_job.asset_selection is not None
            else None
        )
        asset_check_selection = (
            frozenset(
                check.opt_set_param(remote_job.asset_check_selection, "asset_check_selection")
            )
            if remote_job.asset_check_selection is not None
            else None
        )

        execution_plan_snapshot_or_error = sync_get_external_execution_plan_grpc(
            api_client=self.client,
            job_origin=remote_job.get_remote_origin(),
            run_config=run_config,
            job_snapshot_id=remote_job.identifying_job_snapshot_id,
            asset_selection=asset_selection,
            asset_check_selection=asset_check_selection,
            op_selection=remote_job.op_selection,
            step_keys_to_execute=step_keys_to_execute,
            known_state=known_state,
            instance=instance,
        )

        return RemoteExecutionPlan(execution_plan_snapshot=execution_plan_snapshot_or_error)

    @checked
    async def gen_execution_plan(
        self,
        remote_job: RemoteJob,
        run_config: Mapping[str, Any],
        step_keys_to_execute: Sequence[str] | None,
        known_state: KnownExecutionState | None,
        instance: DagsterInstance | None = None,
    ) -> RemoteExecutionPlan:
        from dagster._api.snapshot_execution_plan import gen_external_execution_plan_grpc

        asset_selection = (
            frozenset(check.opt_set_param(remote_job.asset_selection, "asset_selection"))
            if remote_job.asset_selection is not None
            else None
        )
        asset_check_selection = (
            frozenset(
                check.opt_set_param(remote_job.asset_check_selection, "asset_check_selection")
            )
            if remote_job.asset_check_selection is not None
            else None
        )

        execution_plan_snapshot_or_error = await gen_external_execution_plan_grpc(
            api_client=self.client,
            job_origin=remote_job.get_remote_origin(),
            run_config=run_config,
            job_snapshot_id=remote_job.identifying_job_snapshot_id,
            asset_selection=asset_selection,
            asset_check_selection=asset_check_selection,
            op_selection=remote_job.op_selection,
            step_keys_to_execute=step_keys_to_execute,
            known_state=known_state,
            instance=instance,
        )

        return RemoteExecutionPlan(execution_plan_snapshot=execution_plan_snapshot_or_error)

    def _get_subset_remote_job_result(
        self, selector: JobSubsetSelector, get_full_job: Callable[[JobSubsetSelector], RemoteJob]
    ) -> RemoteJobSubsetResult:
        from dagster._api.snapshot_job import sync_get_external_job_subset_grpc

        check.inst_param(selector, "selector", JobSubsetSelector)
        check.invariant(
            selector.location_name == self.name,
            f"PipelineSelector location_name mismatch, got {selector.location_name} expected"
            f" {self.name}",
        )

        remote_repository = self.get_repository(selector.repository_name)
        job_handle = JobHandle(selector.job_name, remote_repository.handle)
        subset = sync_get_external_job_subset_grpc(
            self.client,
            job_handle.get_remote_origin(),
            include_parent_snapshot=False,
            op_selection=selector.op_selection,
            asset_selection=selector.asset_selection,
            asset_check_selection=selector.asset_check_selection,
        )
        # Omit the parent job snapshot for __ASSET_JOB, since it is potentialy very large
        # and unlikely to be useful (unlike subset selections of other jobs)
        if subset.job_data_snap and not is_implicit_asset_job_name(selector.job_name):
            full_job = get_full_job(selector)
            subset = copy(
                subset,
                job_data_snap=copy(subset.job_data_snap, parent_job=full_job.job_snapshot),
            )

        return subset

    async def _gen_subset_remote_job_result(
        self, selector: JobSubsetSelector, get_full_job: Callable[[JobSubsetSelector], RemoteJob]
    ) -> "RemoteJobSubsetResult":
        from dagster._api.snapshot_job import gen_external_job_subset_grpc

        check.inst_param(selector, "selector", JobSubsetSelector)
        check.invariant(
            selector.location_name == self.name,
            f"PipelineSelector location_name mismatch, got {selector.location_name} expected"
            f" {self.name}",
        )

        remote_repository = self.get_repository(selector.repository_name)
        job_handle = JobHandle(selector.job_name, remote_repository.handle)
        subset = await gen_external_job_subset_grpc(
            self.client,
            job_handle.get_remote_origin(),
            include_parent_snapshot=False,
            op_selection=selector.op_selection,
            asset_selection=selector.asset_selection,
            asset_check_selection=selector.asset_check_selection,
        )
        if subset.job_data_snap:
            full_job = get_full_job(selector)
            subset = copy(
                subset,
                job_data_snap=copy(subset.job_data_snap, parent_job=full_job.job_snapshot),
            )

        return subset

    def get_partition_config(
        self,
        repository_handle: RepositoryHandle,
        job_name: str,
        partition_name: str,
        instance: DagsterInstance,
    ) -> "PartitionConfigSnap":
        from dagster._api.snapshot_partition import sync_get_external_partition_config_grpc

        check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
        check.str_param(job_name, "job_name")
        check.str_param(partition_name, "partition_name")

        return sync_get_external_partition_config_grpc(
            self.client, repository_handle, job_name, partition_name, instance
        )

    def get_partition_tags_from_repo(
        self,
        repository_handle: RepositoryHandle,
        job_name: str,
        partition_name: str,
        instance: DagsterInstance,
    ) -> "PartitionTagsSnap":
        from dagster._api.snapshot_partition import sync_get_external_partition_tags_grpc

        check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
        check.str_param(job_name, "job_name")
        check.str_param(partition_name, "partition_name")

        return sync_get_external_partition_tags_grpc(
            self.client, repository_handle, job_name, partition_name, instance
        )

    def get_partition_names_from_repo(
        self, repository_handle: RepositoryHandle, job_name: str
    ) -> Union[PartitionNamesSnap, "PartitionExecutionErrorSnap"]:
        from dagster._api.snapshot_partition import sync_get_external_partition_names_grpc

        return sync_get_external_partition_names_grpc(self.client, repository_handle, job_name)

    def get_schedule_execution_data(
        self,
        instance: DagsterInstance,
        repository_handle: RepositoryHandle,
        schedule_name: str,
        scheduled_execution_time: TimestampWithTimezone | None,
        log_key: Sequence[str] | None,
    ) -> "ScheduleExecutionData":
        from dagster._api.snapshot_schedule import sync_get_external_schedule_execution_data_grpc

        check.inst_param(instance, "instance", DagsterInstance)
        check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
        check.str_param(schedule_name, "schedule_name")
        check.opt_inst_param(
            scheduled_execution_time, "scheduled_execution_time", TimestampWithTimezone
        )
        check.opt_list_param(log_key, "log_key", of_type=str)

        return sync_get_external_schedule_execution_data_grpc(
            self.client,
            instance,
            repository_handle,
            schedule_name,
            scheduled_execution_time,
            log_key,
        )

    def get_sensor_execution_data(
        self,
        instance: DagsterInstance,
        repository_handle: RepositoryHandle,
        name: str,
        last_tick_completion_time: float | None,
        last_run_key: str | None,
        cursor: str | None,
        log_key: Sequence[str] | None,
        last_sensor_start_time: float | None,
    ) -> "SensorExecutionData":
        from dagster._api.snapshot_sensor import sync_get_external_sensor_execution_data_grpc

        return sync_get_external_sensor_execution_data_grpc(
            self.client,
            instance,
            repository_handle,
            name,
            last_tick_completion_time,
            last_run_key,
            cursor,
            log_key,
            last_sensor_start_time,
        )

    def get_partition_set_execution_params(
        self,
        repository_handle: RepositoryHandle,
        partition_set_name: str,
        partition_names: Sequence[str],
        instance: DagsterInstance,
    ) -> "PartitionSetExecutionParamSnap":
        from dagster._api.snapshot_partition import (
            sync_get_external_partition_set_execution_param_data_grpc,
        )

        check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
        check.str_param(partition_set_name, "partition_set_name")
        check.sequence_param(partition_names, "partition_names", of_type=str)

        return sync_get_external_partition_set_execution_param_data_grpc(
            self.client,
            repository_handle,
            partition_set_name,
            partition_names,
            instance,
        )

    def get_notebook_data(self, notebook_path: str) -> bytes:
        from dagster._api.notebook_data import sync_get_streaming_external_notebook_data_grpc

        check.str_param(notebook_path, "notebook_path")
        return sync_get_streaming_external_notebook_data_grpc(self.client, notebook_path)

    def get_dagster_library_versions(self) -> Mapping[str, str] | None:
        return self._dagster_library_versions


def is_implicit_asset_job_name(job_name: str) -> bool:
    return job_name.startswith(IMPLICIT_ASSET_JOB_NAME)



import textwrap

from dagster._core.definitions.decorators.job_decorator import job
from dagster.components.lib.shim_components.base import ShimScaffolder
from dagster.components.scaffold.scaffold import ScaffoldRequest, scaffold_with


class JobScaffolder(ShimScaffolder):
    def get_text(self, request: ScaffoldRequest) -> str:
        return textwrap.dedent(
            f"""\
            import dagster as dg


            @dg.job
            def {request.target_path.stem}():
                pass
            """
        )


scaffold_with(JobScaffolder)(job)



import os


def is_likely_venv_executable(executable_path: str) -> bool:
    # venvs have a file called activate in the same bin directory as the Python executable
    return os.path.exists(os.path.join(os.path.dirname(executable_path), "activate"))



import hashlib
import json
from collections import defaultdict
from collections.abc import Sequence

from dagster._core.errors import DagsterInvalidDefinitionError

# In the Dagster UI users can select partition ranges following the format '2022-01-13...2022-01-14'
# "..." is an invalid substring in partition keys
# The other escape characters are characters that may not display in the Dagster UI.
INVALID_PARTITION_SUBSTRINGS = ["...", "\a", "\b", "\f", "\n", "\r", "\t", "\v", "\0"]


def raise_error_on_invalid_partition_key_substring(partition_keys: Sequence[str]) -> None:
    for partition_key in partition_keys:
        found_invalid_substrs = [
            invalid_substr
            for invalid_substr in INVALID_PARTITION_SUBSTRINGS
            if invalid_substr in partition_key
        ]
        if found_invalid_substrs:
            raise DagsterInvalidDefinitionError(
                f"{found_invalid_substrs} are invalid substrings in a partition key"
            )


def raise_error_on_duplicate_partition_keys(partition_keys: Sequence[str]) -> None:
    counts: dict[str, int] = defaultdict(lambda: 0)
    for partition_key in partition_keys:
        counts[partition_key] += 1
    found_duplicates = [key for key in counts.keys() if counts[key] > 1]
    if found_duplicates:
        raise DagsterInvalidDefinitionError(
            "Partition keys must be unique. Duplicate instances of partition keys:"
            f" {found_duplicates}."
        )


def generate_partition_key_based_definition_id(partition_keys: Sequence[str]) -> str:
    return hashlib.sha1(json.dumps(partition_keys).encode("utf-8")).hexdigest()



import inspect
from collections.abc import Callable, Sequence
from types import ModuleType
from typing import NamedTuple

from dagster import DagsterInvariantViolationError, GraphDefinition, RepositoryDefinition
from dagster._core.code_pointer import load_python_file, load_python_module
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.module_loaders.load_assets_from_modules import (
    load_assets_from_modules,
)

LOAD_ALL_ASSETS = "<<LOAD_ALL_ASSETS>>"


class LoadableTarget(NamedTuple):
    attribute: str
    target_definition: object


def loadable_targets_from_python_file(
    python_file: str, working_directory: str | None = None
) -> Sequence[LoadableTarget]:
    loaded_module = load_python_file(python_file, working_directory)
    return loadable_targets_from_loaded_module(loaded_module)


def loadable_targets_from_python_module(
    module_name: str,
    working_directory: str | None,
    remove_from_path_fn: Callable[[], Sequence[str]] | None = None,
) -> Sequence[LoadableTarget]:
    module = load_python_module(
        module_name,
        working_directory=working_directory,
        remove_from_path_fn=remove_from_path_fn,
    )
    return loadable_targets_from_loaded_module(module)


def loadable_targets_from_python_package(
    package_name: str,
    working_directory: str | None,
    remove_from_path_fn: Callable[[], Sequence[str]] | None = None,
) -> Sequence[LoadableTarget]:
    module = load_python_module(
        package_name, working_directory, remove_from_path_fn=remove_from_path_fn
    )
    return loadable_targets_from_loaded_module(module)


def _format_loadable_def(module: ModuleType, loadable_target: LoadableTarget) -> str:
    return f"{module.__name__}.{loadable_target.attribute}"


def loadable_targets_from_loaded_module(module: ModuleType) -> Sequence[LoadableTarget]:
    from dagster._core.definitions import JobDefinition
    from dagster._utils.test.definitions import LazyDefinitions

    loadable_def_loaders = _loadable_targets_of_type(module, LazyDefinitions)

    if loadable_def_loaders:
        if len(loadable_def_loaders) > 1:
            raise DagsterInvariantViolationError(
                "Cannot have more than one function decorated with @lazy_definitions defined in module scope"
            )

        return loadable_def_loaders

    loadable_defs = _loadable_targets_of_type(module, Definitions)

    if loadable_defs:
        if len(loadable_defs) > 1:
            loadable_def_names = ", ".join(
                _format_loadable_def(module, loadable_def) for loadable_def in loadable_defs
            )
            raise DagsterInvariantViolationError(
                "Cannot have more than one Definitions object defined at module scope."
                f" Found Definitions objects: {loadable_def_names}"
            )

        return loadable_defs

    loadable_repos = _loadable_targets_of_type(module, RepositoryDefinition)
    if loadable_repos:
        return loadable_repos

    loadable_jobs = _loadable_targets_of_type(module, JobDefinition)
    loadable_jobs = _loadable_targets_of_type(module, JobDefinition)

    if len(loadable_jobs) == 1:
        return loadable_jobs

    elif len(loadable_jobs) > 1:
        target_type = "job" if len(loadable_jobs) > 1 else "pipeline"
        raise DagsterInvariantViolationError(
            f'No repository and more than one {target_type} found in "{module.__name__}". If you'
            f" load a file or module directly it must have only one {target_type} in scope."
            f" Found {target_type}s defined in variables or decorated functions:"
            f" {[p.attribute for p in loadable_jobs]!r}."
        )

    loadable_graphs = _loadable_targets_of_type(module, GraphDefinition)

    if len(loadable_graphs) == 1:
        return loadable_graphs

    elif len(loadable_graphs) > 1:
        raise DagsterInvariantViolationError(
            f'More than one graph found in "{module.__name__}". '
            "If you load a file or module directly and it has no repositories, jobs, or "
            "pipelines in scope, it must have no more than one graph in scope. "
            f"Found graphs defined in variables or decorated functions: {[g.attribute for g in loadable_graphs]!r}."
        )

    assets = load_assets_from_modules([module])
    if len(assets) > 0:
        return [LoadableTarget(LOAD_ALL_ASSETS, assets)]

    raise DagsterInvariantViolationError(
        "No Definitions, RepositoryDefinition, Job, Pipeline, Graph, or AssetsDefinition found in "
        f'"{module.__name__}".'
    )


def _loadable_targets_of_type(
    module: ModuleType, klass: type | tuple[type, ...]
) -> Sequence[LoadableTarget]:
    loadable_targets = []
    for name, value in inspect.getmembers(module):
        if isinstance(value, klass):
            loadable_targets.append(LoadableTarget(name, value))

    return loadable_targets


def autodefs_module_target(autoload_defs_module_name: str, working_directory: str | None):
    from dagster.components import load_defs

    module = load_python_module(autoload_defs_module_name, working_directory)
    defs = load_defs(module)
    return LoadableTarget(
        attribute="",
        target_definition=defs,
    )



import textwrap

from dagster_shared.serdes.objects.package_entry import (
    ComponentFeatureData,
    EnvRegistryKey,
    EnvRegistryObjectSnap,
    ScaffoldTargetTypeData,
)

from dagster.components.component.component import Component
from dagster.components.scaffold.scaffold import Scaffolder, get_scaffolder


def _clean_docstring(docstring: str) -> str:
    lines = docstring.strip().splitlines()
    first_line = lines[0]
    if len(lines) == 1:
        return first_line
    else:
        rest = textwrap.dedent("\n".join(lines[1:]))
        return f"{first_line}\n{rest}"


def _get_summary_and_description(obj: object) -> tuple[str | None, str | None]:
    docstring = obj.__doc__
    clean_docstring = _clean_docstring(docstring) if docstring else None
    summary = clean_docstring.split("\n\n")[0] if clean_docstring else None
    description = clean_docstring if clean_docstring else None
    return summary, description


def _get_component_type_data(obj: type[Component]) -> ComponentFeatureData:
    component_schema = obj.get_model_cls()
    component_json_schema = component_schema.model_json_schema() if component_schema else None
    return ComponentFeatureData(schema=component_json_schema)


def _get_scaffold_target_type_data(scaffolder: Scaffolder) -> ScaffoldTargetTypeData:
    scaffolder_schema = scaffolder.get_scaffold_params()
    return ScaffoldTargetTypeData(
        schema=scaffolder_schema.model_json_schema() if scaffolder_schema else None
    )


def get_package_entry_snap(key: EnvRegistryKey, obj: object) -> EnvRegistryObjectSnap:
    type_data = []
    owners = []
    tags = []
    aliases = []
    if isinstance(obj, type) and issubclass(obj, Component):
        type_data.append(_get_component_type_data(obj))
        spec = obj.get_spec()
        owners = spec.owners
        tags = spec.tags
        if obj.__module__ != key.namespace:
            # add the module that defines the class as an alias
            aliases.append(EnvRegistryKey(namespace=obj.__module__, name=obj.__name__))
    scaffolder = get_scaffolder(obj)
    if isinstance(scaffolder, Scaffolder):
        type_data.append(_get_scaffold_target_type_data(scaffolder))
    summary, description = _get_summary_and_description(obj)
    return EnvRegistryObjectSnap(
        key=key,
        aliases=aliases,
        summary=summary,
        owners=owners,
        tags=tags,
        description=description,
        feature_data=type_data,
    )
